# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint:disable=protected-access,no-else-return,unreachable

from copy import deepcopy
from functools import reduce
from typing import List, Optional, Union

from azure.ai.ml._exception_helper import log_and_raise_error
from azure.ai.ml._restclient.v2022_10_01_preview.models import AcrDetails as RestAcrDetails
from azure.ai.ml._restclient.v2022_10_01_preview.models import ArmResourceId as RestArmResourceId
from azure.ai.ml._restclient.v2022_10_01_preview.models import RegistryRegionArmDetails as RestRegistryRegionArmDetails
from azure.ai.ml._restclient.v2022_10_01_preview.models import StorageAccountDetails as RestStorageAccountDetails
from azure.ai.ml._restclient.v2022_10_01_preview.models import SystemCreatedAcrAccount as RestSystemCreatedAcrAccount
from azure.ai.ml._restclient.v2022_10_01_preview.models import (
    SystemCreatedStorageAccount as RestSystemCreatedStorageAccount,
)
from azure.ai.ml._restclient.v2022_10_01_preview.models import UserCreatedAcrAccount as RestUserCreatedAcrAccount
from azure.ai.ml.constants._registry import StorageAccountType
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException

from .util import _make_rest_user_storage_from_id


# This exists despite not being used by the schema validator because this entire
# class is an output only value from the API.
class SystemCreatedAcrAccount:
    def __init__(
        self,
        *,
        acr_account_sku: str,
        arm_resource_id: Optional[str] = None,
    ):
        """Azure ML ACR account.

        :param acr_account_sku: The storage account service tier. Currently
            only Premium is a valid option for registries.
        :type acr_account_sku: str
        :param arm_resource_id: Resource ID of the ACR account.
        :type arm_resource_id: str. Default value is None.
        """
        self.acr_account_sku = acr_account_sku
        self.arm_resource_id = arm_resource_id

    # acr should technically be a union between str and SystemCreatedAcrAccount,
    # but python doesn't accept self class references apparently.
    # Class method instead of normal function to accept possible
    # string input.
    @classmethod
    def _to_rest_object(cls, acr: Union[str, "SystemCreatedAcrAccount"]) -> RestAcrDetails:
        if hasattr(acr, "acr_account_sku") and acr.acr_account_sku is not None:
            # SKU enum requires input to be a capitalized word,
            # so we format the input to be acceptable as long as spelling is
            # correct.
            acr_account_sku = acr.acr_account_sku.capitalize()
            # We DO NOT want to set the arm_resource_id. The backend provides very
            # unhelpful errors if you provide an empty/null/invalid resource ID,
            # and ignores the value otherwise. It's better to avoid setting it in
            # the conversion in this direction at all.
            return RestAcrDetails(
                system_created_acr_account=RestSystemCreatedAcrAccount(
                    acr_account_sku=acr_account_sku,
                )
            )
        else:
            return RestAcrDetails(
                user_created_acr_account=RestUserCreatedAcrAccount(arm_resource_id=RestArmResourceId(resource_id=acr))
            )

    @classmethod
    def _from_rest_object(cls, rest_obj: RestAcrDetails) -> Optional["Union[str, SystemCreatedAcrAccount]"]:
        if not rest_obj:
            return None
        if hasattr(rest_obj, "system_created_acr_account") and rest_obj.system_created_acr_account is not None:
            resource_id = None
            if rest_obj.system_created_acr_account.arm_resource_id:
                resource_id = rest_obj.system_created_acr_account.arm_resource_id.resource_id
            return SystemCreatedAcrAccount(
                acr_account_sku=rest_obj.system_created_acr_account.acr_account_sku,
                arm_resource_id=resource_id,
            )
        elif hasattr(rest_obj, "user_created_acr_account") and rest_obj.user_created_acr_account is not None:
            res: Optional[str] = rest_obj.user_created_acr_account.arm_resource_id.resource_id
            return res
        else:
            return None


class SystemCreatedStorageAccount:
    def __init__(
        self,
        *,
        storage_account_hns: bool,
        storage_account_type: Optional[StorageAccountType],
        arm_resource_id: Optional[str] = None,
        replicated_ids: Optional[List[str]] = None,
        replication_count: int = 1,
    ):
        """
        :param arm_resource_id: Resource ID of the storage account.
        :type arm_resource_id: str
        :param storage_account_hns: Whether or not this storage account
            has hierarchical namespaces enabled.
        :type storage_account_hns: bool
        :param storage_account_type: Allowed values: "Standard_LRS",
            "Standard_GRS, "Standard_RAGRS", "Standard_ZRS", "Standard_GZRS",
            "Standard_RAGZRS", "Premium_LRS", "Premium_ZRS"
        :type storage_account_type: StorageAccountType
        :param replication_count: The number of replicas of this storage account
            that should be created. Defaults to 1. Values less than 1 are invalid.
        :type replication_count: int
        :param replicated_ids: If this storage was replicated, then this is a
            list of all storage IDs with these settings for this registry.
            Defaults to none for un-replicated storage accounts.
        :type replicated_ids: List[str]
        """
        self.arm_resource_id = arm_resource_id
        self.storage_account_hns = storage_account_hns
        self.storage_account_type = storage_account_type
        self.replication_count = replication_count
        self.replicated_ids = replicated_ids


# Per-region information for registries.
class RegistryRegionDetails:
    def __init__(
        self,
        *,
        acr_config: Optional[List[Union[str, SystemCreatedAcrAccount]]] = None,
        location: Optional[str] = None,
        storage_config: Optional[Union[List[str], SystemCreatedStorageAccount]] = None,
    ):
        """Details for each region a registry is in.

        :param acr_details: List of ACR account details. Each value can either be a
            single string representing the arm_resource_id of a user-created
            acr_details object, or a entire SystemCreatedAcrAccount object.
        :type acr_details: List[Union[str, SystemCreatedAcrAccount]]
        :param location: The location where the registry exists.
        :type location: str
        :param storage_account_details: List of storage accounts. Each value
            can either be a single string representing the arm_resource_id of
            a user-created storage account, or an entire
            SystemCreatedStorageAccount object.
        :type storage_account_details: Union[List[str], SystemCreatedStorageAccount]
        """
        self.acr_config = acr_config
        self.location = location
        self.storage_config = storage_config

    @classmethod
    def _from_rest_object(cls, rest_obj: RestRegistryRegionArmDetails) -> Optional["RegistryRegionDetails"]:
        if not rest_obj:
            return None
        converted_acr_details = []
        if rest_obj.acr_details:
            converted_acr_details = [SystemCreatedAcrAccount._from_rest_object(acr) for acr in rest_obj.acr_details]
        storages: Optional[Union[List[str], SystemCreatedStorageAccount]] = []
        if rest_obj.storage_account_details:
            storages = cls._storage_config_from_rest_object(rest_obj.storage_account_details)

        return RegistryRegionDetails(
            acr_config=converted_acr_details,  # type: ignore[arg-type]
            location=rest_obj.location,
            storage_config=storages,
        )

    def _to_rest_object(self) -> RestRegistryRegionArmDetails:
        converted_acr_details = []
        if self.acr_config:
            converted_acr_details = [SystemCreatedAcrAccount._to_rest_object(acr) for acr in self.acr_config]
        storages = []
        if self.storage_config:
            storages = self._storage_config_to_rest_object()
        return RestRegistryRegionArmDetails(
            acr_details=converted_acr_details,
            location=self.location,
            storage_account_details=storages,
        )

    def _storage_config_to_rest_object(self) -> List[RestStorageAccountDetails]:
        storage = self.storage_config
        # storage_config can either be a single system-created storage account,
        # or list of user-inputted id's.
        if (
            storage is not None
            and not isinstance(storage, list)
            and hasattr(storage, "storage_account_type")
            and storage.storage_account_type is not None
        ):
            # We DO NOT want to set the arm_resource_id. The backend provides very
            # unhelpful errors if you provide an empty/null/invalid resource ID,
            # and ignores the value otherwise. It's better to avoid setting it in
            # the conversion in this direction at all.
            # We don't bother processing storage_account_type because the
            # rest version is case insensitive.
            account = RestStorageAccountDetails(
                system_created_storage_account=RestSystemCreatedStorageAccount(
                    storage_account_hns_enabled=storage.storage_account_hns,
                    storage_account_type=storage.storage_account_type,
                )
            )
            # duplicate this value based on the replication_count
            count = storage.replication_count
            if count < 1:
                raise ValueError(f"Replication count cannot be less than 1. Value was: {count}.")
            return [deepcopy(account) for _ in range(0, count)]
        elif storage is not None and not isinstance(storage, SystemCreatedStorageAccount) and len(storage) > 0:
            return [_make_rest_user_storage_from_id(user_id=user_id) for user_id in storage]
        else:
            return []

    @classmethod
    def _storage_config_from_rest_object(
        cls, rest_configs: Optional[List]
    ) -> Optional[Union[List[str], SystemCreatedStorageAccount]]:
        if not rest_configs:
            return None
        num_configs = len(rest_configs)
        if num_configs == 0:
            return None
        system_created_count = reduce(
            # TODO: Bug Item number: 2883323
            lambda x, y: int(x) + int(y),  # type: ignore
            [
                hasattr(config, "system_created_storage_account") and config.system_created_storage_account is not None
                for config in rest_configs
            ],
        )
        # configs should be mono-typed. Either they're all system created
        # or all user created.
        if system_created_count == num_configs:
            # System created case - assume all elements are duplicates
            # of a single storage configuration.
            # Convert back into a single local representation by
            # combining id's into a list, and using the first element's
            # account type and hns.
            first_config = rest_configs[0].system_created_storage_account
            resource_id = None
            if first_config.arm_resource_id:
                resource_id = first_config.arm_resource_id.resource_id
            # account for ids of duplicated if they exist
            replicated_ids = None
            if num_configs > 1:
                replicated_ids = [
                    config.system_created_storage_account.arm_resource_id.resource_id for config in rest_configs
                ]
            return SystemCreatedStorageAccount(
                storage_account_hns=first_config.storage_account_hns_enabled,
                storage_account_type=(
                    (StorageAccountType(first_config.storage_account_type.lower()))
                    if first_config.storage_account_type
                    else None
                ),
                arm_resource_id=resource_id,
                replication_count=num_configs,
                replicated_ids=replicated_ids,
            )
        elif system_created_count == 0:
            return [config.user_created_storage_account.arm_resource_id.resource_id for config in rest_configs]
        else:
            msg = f"""tried reading in a registry whose storage accounts were not
                mono-managed or user-created. {system_created_count} out of {num_configs} were managed."""
            err = ValidationException(
                message=msg,
                target=ErrorTarget.REGISTRY,
                no_personal_data_message=msg,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )
            log_and_raise_error(err)
            return None  # pylint: disable=unreachable



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import List, Optional

from azure.ai.ml._restclient.v2024_10_01_preview.models import IPRule as RestIPRule
from azure.ai.ml._restclient.v2024_10_01_preview.models import NetworkAcls as RestNetworkAcls
from azure.ai.ml.entities._mixins import RestTranslatableMixin


class IPRule(RestTranslatableMixin):
    """Represents an IP rule with a value.

    :param value: An IPv4 address or range in CIDR notation.
    :type value: str
    """

    def __init__(self, value: Optional[str]):
        self.value = value

    def __repr__(self):
        return f"IPRule(value={self.value})"

    def _to_rest_object(self) -> RestIPRule:
        return RestIPRule(value=self.value)

    @classmethod
    def _from_rest_object(cls, obj: RestIPRule) -> "IPRule":
        return cls(value=obj.value)


class DefaultActionType:
    """Specifies the default action when no IP rules are matched."""

    DENY = "Deny"
    ALLOW = "Allow"


class NetworkAcls(RestTranslatableMixin):
    """Network Access Setting for Workspace

    :param default_action: Specifies the default action when no IP rules are matched.
    :type default_action: str
    :param ip_rules: Rules governing the accessibility of a resource from a specific IP address or IP range.
    :type ip_rules: Optional[List[IPRule]]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_workspace.py
                :start-after: [START workspace_network_access_settings]
                :end-before: [END workspace_network_access_settings]
                :language: python
                :dedent: 8
                :caption: Configuring one of the three public network access settings.
    """

    def __init__(
        self,
        *,
        default_action: str = DefaultActionType.ALLOW,
        ip_rules: Optional[List[IPRule]] = None,
    ):
        self.default_action = default_action
        self.ip_rules = ip_rules if ip_rules is not None else []

    def __repr__(self):
        ip_rules_repr = ", ".join(repr(ip_rule) for ip_rule in self.ip_rules)
        return f"NetworkAcls(default_action={self.default_action}, ip_rules=[{ip_rules_repr}])"

    def _to_rest_object(self) -> RestNetworkAcls:
        return RestNetworkAcls(
            default_action=self.default_action,
            ip_rules=(
                [ip_rule._to_rest_object() for ip_rule in self.ip_rules]  # pylint: disable=protected-access
                if self.ip_rules
                else None
            ),
        )

    @classmethod
    def _from_rest_object(cls, obj: RestNetworkAcls) -> "NetworkAcls":
        return cls(
            default_action=obj.default_action,
            ip_rules=(
                [IPRule._from_rest_object(ip_rule) for ip_rule in obj.ip_rules]  # pylint: disable=protected-access
                if obj.ip_rules
                else []
            ),
        )



# pylint: disable=too-many-lines
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
from typing import Any, AsyncIterable, Callable, Dict, Optional, TypeVar

from azure.core.async_paging import AsyncItemPaged, AsyncList
from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
    map_error,
)
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import AsyncHttpResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.mgmt.core.exceptions import ARMErrorFormat

from ... import models as _models
from ..._vendor import _convert_request
from ...operations._operations import build_list_request

T = TypeVar("T")
ClsType = Optional[Callable[[PipelineResponse[HttpRequest, AsyncHttpResponse], T, Dict[str, Any]], Any]]


class Operations:
    """Operations async operations.

    You should not instantiate this class directly. Instead, you should create a Client instance that
    instantiates it for you and attaches it as an attribute.

    :ivar models: Alias to model classes used in this operation group.
    :type models: ~azure.mgmt.machinelearningservices.models
    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = _models

    def __init__(self, client, config, serializer, deserializer) -> None:
        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer
        self._config = config

    @distributed_trace
    def list(self, **kwargs: Any) -> AsyncIterable["_models.AmlOperationListResult"]:
        """Lists all of the available Azure Machine Learning Services REST API operations.

        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: An iterator like instance of either AmlOperationListResult or the result of
         cls(response)
        :rtype:
         ~azure.core.async_paging.AsyncItemPaged[~azure.mgmt.machinelearningservices.models.AmlOperationListResult]
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        api_version = kwargs.pop("api_version", "2023-02-01-preview")  # type: str

        cls = kwargs.pop("cls", None)  # type: ClsType["_models.AmlOperationListResult"]
        error_map = {401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError}
        error_map.update(kwargs.pop("error_map", {}))

        def prepare_request(next_link=None):
            if not next_link:

                request = build_list_request(
                    api_version=api_version,
                    template_url=self.list.metadata["url"],
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)

            else:

                request = build_list_request(
                    api_version=api_version,
                    template_url=next_link,
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)
                request.method = "GET"
            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize("AmlOperationListResult", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)
            return None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            pipeline_response = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=False, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
                raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    list.metadata = {"url": "/providers/Microsoft.MachineLearningServices/operations"}  # type: ignore



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from marshmallow import fields, post_load
from azure.ai.ml._schema.core.fields import NestedField
from azure.ai.ml._schema.core.schema import PathAwareSchema
from .route import RouteSchema


module_logger = logging.getLogger(__name__)


class OnlineInferenceConfigurationSchema(PathAwareSchema):
    liveness_route = NestedField(RouteSchema)
    readiness_route = NestedField(RouteSchema)
    scoring_route = NestedField(RouteSchema)
    entry_script = fields.Str()
    configuration = fields.Dict()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets._artifacts._package.inferencing_server import (
            OnlineInferenceConfiguration,
        )

        return OnlineInferenceConfiguration(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from abc import abstractmethod
from os import PathLike
from pathlib import Path, PurePosixPath
from typing import Any, Dict, Optional, Union
from urllib.parse import urljoin

from azure.ai.ml._utils.utils import is_mlflow_uri, is_url
from azure.ai.ml.entities._assets.asset import Asset


class ArtifactStorageInfo:
    def __init__(
        self,
        name: str,
        version: str,
        relative_path: str,
        datastore_arm_id: Optional[str],
        container_name: str,
        storage_account_url: Optional[str] = None,
        is_file: Optional[bool] = None,
        indicator_file: Optional[str] = None,
    ):
        self.name = name
        self.version = version
        self.relative_path = relative_path
        self.datastore_arm_id = datastore_arm_id
        self.container_name = container_name
        self.storage_account_url = storage_account_url
        self.is_file = is_file
        self.indicator_file = indicator_file

    @property
    def full_storage_path(self) -> Optional[str]:
        if self.storage_account_url is None:
            return f"{self.container_name}/{self.relative_path}"
        return urljoin(self.storage_account_url, f"{self.container_name}/{self.relative_path}")

    @property
    def subdir_path(self) -> Optional[str]:
        if self.is_file:
            path = PurePosixPath(self.relative_path).parent
            if self.storage_account_url is None:
                return f"{self.container_name}/{path}"
            return urljoin(self.storage_account_url, f"{self.container_name}/{path}")
        return self.full_storage_path


class Artifact(Asset):
    """Base class for artifact, can't be instantiated directly.

    :param name: Name of the resource.
    :type name: str
    :param version: Version of the resource.
    :type version: str
    :param path: The local or remote path to the asset.
    :type path: Union[str, os.PathLike]
    :param description: Description of the resource.
    :type description: str
    :param tags: Tag dictionary. Tags can be added, removed, and updated.
    :type tags: dict[str, str]
    :param properties: The asset property dictionary.
    :type properties: dict[str, str]
    :param datastore: The datastore to upload the local artifact to.
    :type datastore: str
    :param kwargs: A dictionary of additional configuration parameters.
    :type kwargs: dict
    """

    def __init__(
        self,
        name: Optional[str] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Dict] = None,
        properties: Optional[Dict] = None,
        path: Optional[Union[str, PathLike]] = None,
        datastore: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            name=name,
            version=version,
            description=description,
            tags=tags,
            properties=properties,
            **kwargs,
        )
        self.path = path
        self.datastore = datastore

    @property
    def path(self) -> Optional[Union[str, PathLike]]:
        return self._path

    @path.setter
    def path(self, value: Optional[Union[str, PathLike]]) -> None:
        if not value or is_url(value) or Path(value).is_absolute() or is_mlflow_uri(value):
            self._path = value
        else:
            self._path = Path(self.base_path, value).resolve()

    @abstractmethod
    def _to_dict(self) -> Dict:
        pass

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)  # pylint: disable = unidiomatic-typecheck
            and self.name == other.name
            and self.id == other.id
            and self.version == other.version
            and self.description == other.description
            and self.tags == other.tags
            and self.properties == other.properties
            and self.base_path == other.base_path
            and self._is_anonymous == other._is_anonymous
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    @abstractmethod
    def _update_path(self, asset_artifact: ArtifactStorageInfo) -> None:
        """Updates an an artifact with the remote path of a local upload.

        :param asset_artifact: The asset storage info of the artifact
        :type asset_artifact: ArtifactStorageInfo
        """



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from typing import Any, Dict

from marshmallow import fields, post_load

from azure.ai.ml._restclient.v2023_04_01_preview.models import (
    InstanceSegmentationPrimaryMetrics,
    ObjectDetectionPrimaryMetrics,
    TaskType,
)
from azure.ai.ml._schema.automl.image_vertical.image_model_distribution_settings import (
    ImageModelDistributionSettingsInstanceSegmentationSchema,
    ImageModelDistributionSettingsObjectDetectionSchema,
)
from azure.ai.ml._schema.automl.image_vertical.image_model_settings import ImageModelSettingsObjectDetectionSchema
from azure.ai.ml._schema.automl.image_vertical.image_vertical import ImageVerticalSchema
from azure.ai.ml._schema.core.fields import NestedField, StringTransformedEnum
from azure.ai.ml._utils.utils import camel_to_snake
from azure.ai.ml.constants._job.automl import AutoMLConstants


class ImageObjectDetectionSchema(ImageVerticalSchema):
    task_type = StringTransformedEnum(
        allowed_values=TaskType.IMAGE_OBJECT_DETECTION,
        casing_transform=camel_to_snake,
        data_key=AutoMLConstants.TASK_TYPE_YAML,
        required=True,
    )
    primary_metric = StringTransformedEnum(
        allowed_values=ObjectDetectionPrimaryMetrics.MEAN_AVERAGE_PRECISION,
        casing_transform=camel_to_snake,
        load_default=camel_to_snake(ObjectDetectionPrimaryMetrics.MEAN_AVERAGE_PRECISION),
    )
    training_parameters = NestedField(ImageModelSettingsObjectDetectionSchema())
    search_space = fields.List(NestedField(ImageModelDistributionSettingsObjectDetectionSchema()))

    @post_load
    def make(self, data, **kwargs) -> Dict[str, Any]:
        data.pop("task_type")
        return data


class ImageInstanceSegmentationSchema(ImageVerticalSchema):
    task_type = StringTransformedEnum(
        allowed_values=TaskType.IMAGE_INSTANCE_SEGMENTATION,
        casing_transform=camel_to_snake,
        data_key=AutoMLConstants.TASK_TYPE_YAML,
        required=True,
    )
    primary_metric = StringTransformedEnum(
        allowed_values=[InstanceSegmentationPrimaryMetrics.MEAN_AVERAGE_PRECISION],
        casing_transform=camel_to_snake,
        load_default=camel_to_snake(InstanceSegmentationPrimaryMetrics.MEAN_AVERAGE_PRECISION),
    )
    training_parameters = NestedField(ImageModelSettingsObjectDetectionSchema())
    search_space = fields.List(NestedField(ImageModelDistributionSettingsInstanceSegmentationSchema()))

    @post_load
    def make(self, data, **kwargs) -> Dict[str, Any]:
        data.pop("task_type")
        return data



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------


class SnakeCaseFineTuningTaskTypes:
    CHAT_COMPLETION = "chat_completion"
    TEXT_COMPLETION = "text_completion"
    TEXT_CLASSIFICATION = "text_classification"
    QUESTION_ANSWERING = "question_answering"
    TEXT_SUMMARIZATION = "text_summarization"
    TOKEN_CLASSIFICATION = "token_classification"
    TEXT_TRANSLATION = "text_translation"
    IMAGE_CLASSIFICATION = "image_classification"
    IMAGE_INSTANCE_SEGMENTATION = "image_instance_segmentation"
    IMAGE_OBJECT_DETECTION = "image_object_detection"
    VIDEO_MULTI_OBJECT_TRACKING = "video_multi_object_tracking"



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Optional

from azure.ai.ml._restclient.v2023_06_01_preview.models import ComputeRuntimeDto as RestComputeRuntimeDto
from azure.ai.ml.entities._mixins import RestTranslatableMixin


class ComputeRuntime(RestTranslatableMixin):
    """Spark compute runtime configuration.

    :keyword spark_runtime_version: Spark runtime version.
    :paramtype spark_runtime_version: Optional[str]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_compute.py
            :start-after: [START compute_runtime]
            :end-before: [END compute_runtime]
            :language: python
            :dedent: 8
            :caption: Creating a ComputeRuntime object.
    """

    def __init__(
        self,
        *,
        spark_runtime_version: Optional[str] = None,
    ) -> None:
        self.spark_runtime_version = spark_runtime_version

    def _to_rest_object(self) -> RestComputeRuntimeDto:
        return RestComputeRuntimeDto(spark_runtime_version=self.spark_runtime_version)

    @classmethod
    def _from_rest_object(cls, obj: RestComputeRuntimeDto) -> Optional["ComputeRuntime"]:
        if not obj:
            return None
        return ComputeRuntime(spark_runtime_version=obj.spark_runtime_version)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
from typing import Any

from marshmallow import fields, post_load

from azure.ai.ml._schema import PatchedSchemaMeta

module_logger = logging.getLogger(__name__)


class RequestLoggingSchema(metaclass=PatchedSchemaMeta):
    capture_headers = fields.List(fields.Str())

    # pylint: disable=unused-argument
    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:
        from azure.ai.ml.entities._deployment.request_logging import RequestLogging

        return RequestLogging(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Contains exception module in Azure Machine Learning SDKv2.

This includes enums and classes for exceptions.
"""
import logging
from enum import Enum
from typing import Optional, Union

from azure.core.exceptions import AzureError

module_logger = logging.getLogger(__name__)


class ValidationErrorType(Enum):
    """Error types to be specified when using ValidationException class. Types are then used in raise_error.py to format
    a detailed error message for users.

    When using ValidationException, specify the type that best describes the nature of the error being captured.
    If no type fits, add a new enum here and update raise_error.py to handle it.

    Types of validation errors:

    * INVALID_VALUE -> One or more schema fields are invalid (e.g. incorrect type or format)
    * UNKNOWN_FIELD -> A least one unrecognized schema parameter is specified
    * MISSING_FIELD -> At least one required schema parameter is missing
    * FILE_OR_FOLDER_NOT_FOUND -> One or more files or folder paths do not exist
    * CANNOT_SERIALIZE -> Same as "Cannot dump". One or more fields cannot be serialized by marshmallow.
    * CANNOT_PARSE -> YAML file cannot be parsed
    * RESOURCE_NOT_FOUND -> Resource could not be found
    * GENERIC -> Undefined placeholder. Avoid using.
    """

    INVALID_VALUE = "INVALID VALUE"
    UNKNOWN_FIELD = "UNKNOWN FIELD"
    MISSING_FIELD = "MISSING FIELD"
    FILE_OR_FOLDER_NOT_FOUND = "FILE OR FOLDER NOT FOUND"
    CANNOT_SERIALIZE = "CANNOT DUMP"
    CANNOT_PARSE = "CANNOT PARSE"
    RESOURCE_NOT_FOUND = "RESOURCE NOT FOUND"
    GENERIC = "GENERIC"


class ErrorCategory:
    USER_ERROR = "UserError"
    SYSTEM_ERROR = "SystemError"
    UNKNOWN = "Unknown"


class ErrorTarget:
    BATCH_ENDPOINT = "BatchEndpoint"
    BATCH_DEPLOYMENT = "BatchDeployment"
    LOCAL_ENDPOINT = "LocalEndpoint"
    CODE = "Code"
    COMPONENT = "Component"
    DATA = "Data"
    ENVIRONMENT = "Environment"
    JOB = "Job"
    COMMAND_JOB = "CommandJob"
    SPARK_JOB = "SparkJob"
    DATA_TRANSFER_JOB = "DataTransferJob"
    LOCAL_JOB = "LocalJob"
    MODEL = "Model"
    INDEX = "Index"
    ONLINE_DEPLOYMENT = "OnlineDeployment"
    ONLINE_ENDPOINT = "OnlineEndpoint"
    ASSET = "Asset"
    DATASTORE = "Datastore"
    BLOB_DATASTORE = "BlobDatastore"
    FILE_DATASTORE = "FileDatastore"
    GEN1_DATASTORE = "Gen1Datastore"
    GEN2_DATASTORE = "Gen2Datastore"
    WORKSPACE = "Workspace"
    CAPABILITY_HOST = "CapabilityHost"
    COMPUTE = "Compute"
    DEPLOYMENT = "Deployment"
    ENDPOINT = "Endpoint"
    AUTOML = "AutoML"
    FINETUNING = "FineTuning"
    PIPELINE = "Pipeline"
    SWEEP_JOB = "SweepJob"
    GENERAL = "General"
    IDENTITY = "Identity"
    ARM_DEPLOYMENT = "ArmDeployment"
    ARM_RESOURCE = "ArmResource"
    ARTIFACT = "Artifact"
    SCHEDULE = "Schedule"
    REGISTRY = "Registry"
    UNKNOWN = "Unknown"
    FEATURE_SET = "FeatureSet"
    FEATURE_STORE_ENTITY = "FeatureStoreEntity"
    MODEL_MONITORING = "ModelMonitoring"
    SERVERLESS_ENDPOINT = "ServerlessEndpoint"  ## EXPERIMENTAL/IN PREVIEW


class MlException(AzureError):
    """The base class for all exceptions raised in AzureML SDK code base. If there is a need to define a custom
    exception type, that custom exception type should extend from this class.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data. This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    :param error: The original exception if any.
    :type error: Exception
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        self._error_category = error_category
        self._target = target
        self._no_personal_data_message = no_personal_data_message
        super().__init__(message, *args, **kwargs)

    @property
    def target(self):
        """Return the error target.

        :return: The error target.
        :rtype: ErrorTarget
        """
        return self._target

    @target.setter
    def target(self, value):
        self._target = value

    @property
    def no_personal_data_message(self):
        """Return the error message with no personal data.

        :return: No personal data error message.
        :rtype: str
        """
        return self._no_personal_data_message

    @no_personal_data_message.setter
    def no_personal_data_message(self, value):
        self._no_personal_data_message = value

    @property
    def error_category(self):
        """Return the error category.

        :return: The error category.
        :rtype: ErrorCategory
        """
        return self._error_category

    @error_category.setter
    def error_category(self, value):
        self._error_category = value


class DeploymentException(MlException):
    """Class for all exceptions related to Deployments.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(DeploymentException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class ComponentException(MlException):
    """Class for all exceptions related to Components.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(ComponentException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class JobException(MlException):
    """Class for all exceptions related to Jobs.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(JobException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class ModelException(MlException):
    """Class for all exceptions related to Models.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(ModelException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class AssetException(MlException):
    """Class for all exceptions related to Assets.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(AssetException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class ScheduleException(MlException):
    """Class for all exceptions related to Job Schedules.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(ScheduleException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class ValidationException(MlException):
    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        error_type: ValidationErrorType = ValidationErrorType.GENERIC,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.USER_ERROR,
        **kwargs,
    ):
        """Class for all exceptions raised as part of client-side schema validation.

        :param message: A message describing the error. This is the error message the user will see.
        :type message: str
        :param no_personal_data_message: The error message without any personal data.
            This will be pushed to telemetry logs.
        :type no_personal_data_message: str
        :param error_type: The error type, chosen from one of the values of ValidationErrorType enum class.
        :type error_type: ValidationErrorType
        :param target: The name of the element that caused the exception to be thrown.
        :type target: ErrorTarget
        :param error_category: The error category, defaults to Unknown.
        :type error_category: ErrorCategory
        :param error: The original exception if any.
        :type error: Exception
        """
        super(ValidationException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )

        self.raw_error = message  # used for CLI error formatting

        if error_type in list(ValidationErrorType):
            self._error_type = error_type
        else:
            msg = f"Error type {error_type} is not a member of the ValidationErrorType enum class."
            raise MlException(message=msg, no_personal_data_message=msg)

    @property
    def error_type(self):
        """Return the error type.

        :return: The error type.
        :rtype: ValidationErrorType
        """
        return self._error_type

    @error_type.setter
    def error_type(self, value):
        self._error_type = value


class AssetPathException(MlException):
    """Class for the exception raised when an attempt is made to update the path of an existing asset. Asset paths are
    immutable.

    :param message: A message describing the error. This is the error message the user will see.
    :type message: str
    :param no_personal_data_message: The error message without any personal data.
        This will be pushed to telemetry logs.
    :type no_personal_data_message: str
    :param target: The name of the element that caused the exception to be thrown.
    :type target: ErrorTarget
    :param error_category: The error category, defaults to Unknown.
    :type error_category: ErrorCategory
    """

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        *args,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
        **kwargs,
    ):
        super(AssetPathException, self).__init__(
            message=message,
            target=target,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class EmptyDirectoryError(MlException):
    """Exception raised when an empty directory is provided as input for an I/O operation."""

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        target: ErrorTarget = ErrorTarget.UNKNOWN,
        error_category: ErrorCategory = ErrorCategory.UNKNOWN,
    ):
        self.message = message
        super(EmptyDirectoryError, self).__init__(
            message=self.message,
            no_personal_data_message=no_personal_data_message,
            target=target,
            error_category=error_category,
        )


class UserErrorException(MlException):
    """Exception raised when invalid or unsupported inputs are provided."""

    def __init__(
        self,
        message,
        no_personal_data_message=None,
        error_category=ErrorCategory.USER_ERROR,
        target: ErrorTarget = ErrorTarget.PIPELINE,
    ):
        super().__init__(
            message=message,
            target=target,
            no_personal_data_message=no_personal_data_message,
            error_category=error_category,
        )


class CannotSetAttributeError(UserErrorException):
    """Exception raised when a user try setting attributes of inputs/outputs."""

    def __init__(self, object_name):
        msg = "It is not allowed to set attribute of %r." % object_name
        super(CannotSetAttributeError, self).__init__(
            message=msg,
            no_personal_data_message="It is not allowed to set attribute of object.",
        )


class UnsupportedParameterKindError(UserErrorException):
    """Exception raised when a user try setting attributes of inputs/outputs."""

    def __init__(self, func_name, parameter_kind=None):
        parameter_kind = parameter_kind or "*args or **kwargs"
        msg = "%r: dsl pipeline does not accept %s as parameters." % (func_name, parameter_kind)
        super(UnsupportedParameterKindError, self).__init__(message=msg, no_personal_data_message=msg)


class KeywordError(UserErrorException):
    """Super class of all type keyword error."""

    def __init__(self, message, no_personal_data_message=None):
        super().__init__(message=message, no_personal_data_message=no_personal_data_message)


class UnexpectedKeywordError(KeywordError):
    """Exception raised when an unexpected keyword parameter is provided in dynamic functions."""

    def __init__(self, func_name, keyword, keywords=None):
        message = "%s() got an unexpected keyword argument %r" % (func_name, keyword)
        message += ", valid keywords: %s." % ", ".join("%r" % key for key in keywords) if keywords else "."
        super().__init__(message=message, no_personal_data_message=message)


class UnexpectedAttributeError(KeywordError, AttributeError):
    """Exception raised when an unexpected keyword is invoked by attribute, e.g. inputs.invalid_key."""

    def __init__(self, keyword, keywords=None):
        message = "Got an unexpected attribute %r" % keyword
        message += ", valid attributes: %s." % ", ".join("%r" % key for key in keywords) if keywords else "."
        super().__init__(message=message, no_personal_data_message=message)


class MissingPositionalArgsError(KeywordError):
    """Exception raised when missing positional keyword parameter in dynamic functions."""

    def __init__(self, func_name, missing_args):
        message = "%s() missing %d required positional argument(s): %s." % (
            func_name,
            len(missing_args),
            ", ".join("%r" % key for key in missing_args) if missing_args else ".",
        )
        super().__init__(message=message, no_personal_data_message=message)


class TooManyPositionalArgsError(KeywordError):
    """Exception raised when too many positional arguments is provided in dynamic functions."""

    def __init__(self, func_name, min_number, max_number, given_number):
        message = "%s() takes %s positional argument but %d were given." % (
            func_name,
            min_number if min_number == max_number else f"{min_number} to {max_number}",
            given_number,
        )
        super().__init__(message=message, no_personal_data_message=message)


class MultipleValueError(KeywordError):
    """Exception raised when giving multiple value of a keyword parameter in dynamic functions."""

    def __init__(self, func_name, keyword):
        message = "%s() got multiple values for argument %r." % (func_name, keyword)
        super().__init__(message=message, no_personal_data_message=message)


class ParamValueNotExistsError(KeywordError):
    """Exception raised when items in non_pipeline_inputs not in keyword parameters in dynamic functions."""

    def __init__(self, func_name, keywords):
        message = "%s() got unexpected params in non_pipeline_inputs %r." % (func_name, keywords)
        super().__init__(message=message, no_personal_data_message=message)


class UnsupportedOperationError(UserErrorException):
    """Exception raised when specified operation is not supported."""

    def __init__(self, operation_name):
        message = "Operation %s is not supported." % operation_name
        super().__init__(message=message, no_personal_data_message=message)


class LocalEndpointNotFoundError(MlException):
    """Exception raised if local endpoint cannot be found."""

    def __init__(
        self,
        endpoint_name: str,
        deployment_name: Optional[str] = None,
        error_category=ErrorCategory.USER_ERROR,
    ):
        resource_name = (
            f"Local deployment ({endpoint_name} / {deployment_name})"
            if deployment_name
            else f"Local endpoint ({endpoint_name})"
        )
        err = f"{resource_name} does not exist."
        resource_type = "deployment" if deployment_name else "endpoint"
        super().__init__(
            message=err,
            error_category=error_category,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message=f"Local ({resource_type}) does not exist.",
        )


class LocalEndpointInFailedStateError(MlException):
    """Exception raised when local endpoint is in Failed state."""

    def __init__(self, endpoint_name, deployment_name=None, error_category=ErrorCategory.UNKNOWN):
        resource_name = (
            f"Local deployment ({endpoint_name} / {deployment_name})"
            if deployment_name
            else f"Local endpoint ({endpoint_name})"
        )
        err = f"{resource_name} is in failed state. Try getting logs to debug scoring script."
        resource_type = "deployment" if deployment_name else "endpoint"
        super().__init__(
            message=err,
            error_category=error_category,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message=(
                f"Local ({resource_type}) is in failed state. Try getting logs to debug scoring script."
            ),
        )


class DockerEngineNotAvailableError(MlException):
    """Exception raised when local Docker Engine is unavailable for local operation."""

    def __init__(self, error_category=ErrorCategory.UNKNOWN):
        msg = "Please make sure Docker Engine is installed and running. https://docs.docker.com/engine/install/"
        super().__init__(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.LOCAL_ENDPOINT,
            error_category=error_category,
        )


class MultipleLocalDeploymentsFoundError(MlException):
    """Exception raised when no deployment name is specified for local endpoint even though multiple deployments
    exist."""

    def __init__(self, endpoint_name: str, error_category=ErrorCategory.UNKNOWN):
        super().__init__(
            message=f"Multiple deployments found for local endpoint ({endpoint_name}), please specify deployment name.",
            no_personal_data_message="Multiple deployments found for local endpoint, please specify deployment name.",
            error_category=error_category,
            target=ErrorTarget.LOCAL_ENDPOINT,
        )


class InvalidLocalEndpointError(MlException):
    """Exception raised when local endpoint is invalid."""

    def __init__(
        self,
        message: str,
        no_personal_data_message: str,
        error_category=ErrorCategory.USER_ERROR,
    ):
        super().__init__(
            message=message,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message=no_personal_data_message,
            error_category=error_category,
        )


class LocalEndpointImageBuildError(MlException):
    """Exception raised when local endpoint's Docker image build is unsuccessful."""

    def __init__(self, error: Union[str, Exception], error_category=ErrorCategory.UNKNOWN):
        err = f"Building the local endpoint image failed with error: {str(error)}"
        super().__init__(
            message=err,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message="Building the local endpoint image failed with error.",
            error_category=error_category,
            error=error if error is Exception else None,
        )


class CloudArtifactsNotSupportedError(MlException):
    """Exception raised when remote cloud artifacts are used with local endpoints.

    Local endpoints only support local artifacts.
    """

    def __init__(
        self,
        endpoint_name: str,
        invalid_artifact: str,
        deployment_name: Optional[str] = None,
        error_category=ErrorCategory.USER_ERROR,
    ):
        resource_name = (
            f"local deployment ({endpoint_name} / {deployment_name})"
            if deployment_name
            else f"local endpoint ({endpoint_name})"
        )
        err = (
            "Local endpoints only support local artifacts. '%s' in '%s' referenced cloud artifacts.",
            invalid_artifact,
            resource_name,
        )
        super().__init__(
            message=err,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message="Local endpoints only support local artifacts.",
            error_category=error_category,
        )


class RequiredLocalArtifactsNotFoundError(MlException):
    """Exception raised when local artifact is not provided for local endpoint."""

    def __init__(
        self,
        endpoint_name: str,
        required_artifact: str,
        required_artifact_type: str,
        deployment_name: Optional[str] = None,
        error_category=ErrorCategory.USER_ERROR,
    ):
        resource_name = (
            f"Local deployment ({endpoint_name} / {deployment_name})"
            if deployment_name
            else f"Local endpoint ({endpoint_name})"
        )
        err = (
            (
                "Local endpoints only support local artifacts. '%s' did not contain required local artifact '%s'"
                " of type '%s'."
            ),
            resource_name,
            required_artifact,
            required_artifact_type,
        )
        super().__init__(
            message=err,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message="Resource group did not contain required local artifact.",
            error_category=error_category,
        )


class JobParsingError(MlException):
    """Exception that the job data returned by MFE cannot be parsed."""

    def __init__(self, error_category, no_personal_data_message, message, *args, **kwargs):
        super(JobParsingError, self).__init__(
            message=message,
            target=ErrorTarget.JOB,
            error_category=error_category,
            no_personal_data_message=no_personal_data_message,
            *args,
            **kwargs,
        )


class PipelineChildJobError(MlException):
    """Exception that the pipeline child job is not supported."""

    ERROR_MESSAGE_TEMPLATE = "az ml job {command} is not supported on pipeline child job, {prompt_message}."
    PROMPT_STUDIO_UI_MESSAGE = "please go to studio UI to do related actions{url}"
    PROMPT_PARENT_MESSAGE = "please use this command on pipeline parent job"

    def __init__(self, job_id: str, command: str = "parse", prompt_studio_ui: bool = False):
        from azure.ai.ml.entities._job._studio_url_from_job_id import studio_url_from_job_id

        if prompt_studio_ui:
            url = studio_url_from_job_id(job_id)
            if url:
                url = f": {url}"
            prompt_message = self.PROMPT_STUDIO_UI_MESSAGE.format(url=url)
        else:
            prompt_message = self.PROMPT_PARENT_MESSAGE

        super(PipelineChildJobError, self).__init__(
            message=self.ERROR_MESSAGE_TEMPLATE.format(command=command, prompt_message=prompt_message),
            no_personal_data_message="Pipeline child job is not supported currently.",
            target=ErrorTarget.JOB,
            error_category=ErrorCategory.USER_ERROR,
        )
        self.job_id = job_id


## -------- VSCode Debugger Errors -------- ##


class InvalidVSCodeRequestError(MlException):
    """Exception raised when VS Code Debug is invoked with a remote endpoint.

    VSCode debug is only supported for local endpoints.
    """

    def __init__(self, error_category=ErrorCategory.USER_ERROR, msg=None):
        super().__init__(
            message=msg,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message=msg,
            error_category=error_category,
        )


class VSCodeCommandNotFound(MlException):
    """Exception raised when VSCode instance cannot be instantiated."""

    def __init__(self, output=None, error_category=ErrorCategory.USER_ERROR):
        error_msg = f" due to error: [{output}]" if output else ""
        super().__init__(
            message=(
                f"Could not start VSCode instance{error_msg}. Please make sure the VSCode command "
                "'code' is installed and accessible from PATH environment variable. "
                "See https://code.visualstudio.com/docs/editor/command-line#_common-questions.\n"
            ),
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message="Could not start VSCode instance.",
            error_category=error_category,
        )


class LocalDeploymentGPUNotAvailable(MlException):
    """Exception raised when local_enable_gpu is set and Nvidia GPU is not available."""

    def __init__(self, error_category=ErrorCategory.USER_ERROR, msg=None):
        super().__init__(
            message=msg,
            target=ErrorTarget.LOCAL_ENDPOINT,
            no_personal_data_message=msg,
            error_category=error_category,
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import importlib

from typing import Any, Dict, List, Optional, Union

from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml.entities import CommandComponent, PipelineJob
from azure.ai.ml.entities._assets.federated_learning_silo import FederatedLearningSilo
from azure.ai.ml.entities._builders.fl_scatter_gather import FLScatterGather


def _check_for_import(package_name: str) -> None:
    try:
        importlib.import_module(package_name)
    except ImportError as e:
        raise ImportError(
            "The DSL FL Node has an additional requirement above the rest of the "
            + "AML SDK repo in that the mldesigner package is required. Please run `pip install mldesigner` "
            + "and try again."
        ) from e


@experimental
def fl_scatter_gather(
    *,
    silo_configs: List[FederatedLearningSilo],
    silo_component: Union[PipelineJob, CommandComponent],
    aggregation_component: Union[PipelineJob, CommandComponent],
    aggregation_compute: Optional[str] = None,
    aggregation_datastore: Optional[str] = None,
    shared_silo_kwargs: Optional[Dict] = None,
    aggregation_kwargs: Optional[Dict] = None,
    silo_to_aggregation_argument_map: Optional[Dict] = None,
    aggregation_to_silo_argument_map: Optional[Dict] = None,
    max_iterations: int = 1,
    _create_default_mappings_if_needed: bool = False,
    **kwargs: Any,
) -> FLScatterGather:
    """A federated learning scatter-gather subgraph node.

    It's assumed that this will be used inside of a `@pipeline`-decorated function in order to create a subgraph which
    will:
        - Execute a specified pipeline step multiple times (the silo step), with each execution using slightly
            different inputs, datastores, and computes based on an inputted config.
        - Merge the outputs of the multiple silo steps into a single input for another step (the aggregation step),
            which will then process the values into a single unified result.
        - With the process above being a 'scatter gather' sequence, iterate and perform the scatter gather
            a number of times according to the max_iterations input, with the output of any given iteration's
            aggregation step being fed back into the silo steps of the subsequent iteration.
        - Return the outputs of the last iteration's aggregation step as the node's final output.

    The process of assigning computes, datastores, and inputs to steps is called 'anchoring'. The following
    details of the anchoring process should be noted:
        - Computes will always be overridden to their respective compute.
        - The paths of 'internal' outputs for silo steps (i.e. a component output that becomes an input for another
            component within the silo step) are anchored to that silo's datastore. All other outputs are anchored
            to the aggregation datastore.
        - Some steps are automatically created by this node to merge inputs from the silo steps to the aggregation
            step. These steps are anchored in the same manner as the aggregation step.
        - Datastore anchoring ONLY occurs if an output's path value is empty. If a path already exists, it will NOT
            try to replace it.

    The process of trimming down inputs from the silo steps to a single input for the aggregate step has some
    caveats:
        - Only silo outputs of type mltable and uri_folder are supported.
        - Both the above output types become an mltable which mounts all the silo step outputs as sub-folders.

    The expected use case of this node is shown in the following code snippet:

    .. code-block:: python

        @experimental
        def fl_pipeline():
            fl_node = fl_scatter_gather(**many_inputs)
            return {"pipeline_output" : fl_node.outputs["aggregated_model"]}

        submitted_pipeline_job = my_client.jobs.create_or_update(fl_pipeline(), experiment_name="example_fl_pipeline")

    :keyword silo_configs: A list of FederatedLearningSilo objects,
        which contain the necessary data to reconfigure components to run on specific computes and datastores,
        while also targeting specific inputs located on the aforementioned datastores.
    :paramtype silo_configs: List[~azure.ai.ml.entities._assets.federated_learning_silo.FederatedLearningSilo]
    :keyword silo_component: A pipeline step that will be run multiple times across different silos, as specified
        by the silo_configs input. In a typical horizontal federated learning context, this step is what will perform
        model training using siloed subsets of data. Can be either a PipelineJob or a CommandComponent.
    :paramtype silo_component: Union[~azure.ai.ml.entities.PipelineJob, ~azure.ai.ml.entities.CommandComponent]
    :keyword aggregation_component: A pipeline step which receives inputs from the myriad executed silo components,
        and does something with them. In a typical horizontal federated learning context, this component will merge
        the models that were independently trained on each silo's data in a single model. Can be either a
        PipelineJob or a CommandComponent.
    :paramtype aggregation_component: Union[
        ~azure.ai.ml.entities.PipelineJob,
        ~azure.ai.ml.entities.CommandComponent]
    :keyword aggregation_compute: The name of the compute that the aggregation component will use.
    :paramtype aggregation_compute: str
    :keyword aggregation_datastore: The name of the datastore that the aggregation component will use.
    :paramtype aggregation_datastore: str
    :keyword shared_silo_kwargs: A dictionary of string keywords to component inputs. This dictionary is treated
        like kwargs and is injected into ALL executed silo components.
    :paramtype shared_silo_kwargs: Dict
    :keyword aggregation_kwargs: A dictionary of string keywords to component inputs. This dictionary is treated
        like kwargs and is injected into ALL executed aggregation components.
    :paramtype aggregation_kwargs: Dict
    :keyword silo_to_aggregation_argument_map: A dictionary specifying the mapping of outputs from the silo step to
        inputs in the aggregation step. The keys should be output names from the silo step, and the values should be
        input names in the aggregation step. This allows for customization of the mapping between the steps.
    :paramtype silo_to_aggregation_argument_map: Dict
    :keyword aggregation_to_silo_argument_map: A dictionary specifying the mapping of outputs from the aggregation step
        to inputs in the silo step. The keys should be output names from the aggregation step, and the values should be
        input names in the silo step. This allows for customization of the mapping between the steps.
    :paramtype aggregation_to_silo_argument_map: Dict
    :keyword max_iterations: The maximum number of scatter gather iterations that should be performed.
    :paramtype max_iterations: int
    :keyword _create_default_mappings_if_needed:
        If True, try to automatically create input/output mappings if they're unset.
    :paramtype _create_default_mappings_if_needed: bool
    :return: The federated learning scatter-gather subgraph node.
    :rtype: ~azure.ai.ml.entities._builders.fl_scatter_gather.FLScatterGather

    .. warning::

        Using this node directly from the SDK repo requires that the user have the 'mldesigner' package installed,
        which is not a standard dependency of this package.
    """
    # Private kwargs:
    # _create_default_mappings_if_needed: if true, then try to automatically create i/o mappings if they're unset.

    # check that mldesigner is available
    _check_for_import("mldesigner")

    # Like other DSL nodes, this is just a wrapper around a node builder entity initializer.
    return FLScatterGather(
        silo_configs=silo_configs,
        silo_component=silo_component,  # type: ignore[arg-type]
        aggregation_component=aggregation_component,  # type: ignore[arg-type]
        shared_silo_kwargs=shared_silo_kwargs,
        aggregation_compute=aggregation_compute,
        aggregation_datastore=aggregation_datastore,
        aggregation_kwargs=aggregation_kwargs,
        silo_to_aggregation_argument_map=silo_to_aggregation_argument_map,
        aggregation_to_silo_argument_map=aggregation_to_silo_argument_map,
        max_iterations=max_iterations,
        create_default_mappings_if_needed=_create_default_mappings_if_needed,
        **kwargs,
    )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------


import abc
import os
from os import PathLike
from pathlib import Path
from typing import IO, Any, AnyStr, Dict, List, Optional, Tuple, Union, cast

from msrest import Serializer

from azure.ai.ml._restclient.v2022_10_01 import models
from azure.ai.ml._telemetry.logging_handler import in_jupyter_notebook
from azure.ai.ml._utils.utils import dump_yaml

from ..constants._common import BASE_PATH_CONTEXT_KEY
from ._system_data import SystemData


class Resource(abc.ABC):
    """Base class for entity classes.

    Resource is an abstract object that serves as a base for creating resources. It contains common properties and
    methods for all resources.

    This class should not be instantiated directly. Instead, use one of its subclasses.

    :param name: The name of the resource.
    :type name: str
    :param description: The description of the resource.
    :type description: Optional[str]
    :param tags: Tags can be added, removed, and updated.
    :type tags: Optional[dict]
    :param properties: The resource's property dictionary.
    :type properties: Optional[dict]
    :keyword print_as_yaml: Specifies if the the resource should print out as a YAML-formatted object. If False,
        the resource will print out in a more-compact style. By default, the YAML output is only used in Jupyter
        notebooks. Be aware that some bookkeeping values are shown only in the non-YAML output.
    :paramtype print_as_yaml: bool
    """

    def __init__(
        self,
        name: Optional[str],
        description: Optional[str] = None,
        tags: Optional[Dict] = None,
        properties: Optional[Dict] = None,
        **kwargs: Any,
    ) -> None:
        self.name = name
        self.description = description
        self.tags: Optional[Dict] = dict(tags) if tags else {}
        self.properties = dict(properties) if properties else {}
        # Conditional assignment to prevent entity bloat when unused.
        self._print_as_yaml = kwargs.pop("print_as_yaml", False)

        # Hide read only properties in kwargs
        self._id = kwargs.pop("id", None)
        self.__source_path: Union[str, PathLike] = kwargs.pop("source_path", "")
        self._base_path = kwargs.pop(BASE_PATH_CONTEXT_KEY, None) or os.getcwd()  # base path should never be None
        self._creation_context: Optional[SystemData] = kwargs.pop("creation_context", None)
        client_models = {k: v for k, v in models.__dict__.items() if isinstance(v, type)}
        self._serialize = Serializer(client_models)
        self._serialize.client_side_validation = False
        super().__init__(**kwargs)

    @property
    def _source_path(self) -> Union[str, PathLike]:
        # source path is added to display file location for validation error messages
        # usually, base_path = Path(source_path).parent if source_path else os.getcwd()
        return self.__source_path

    @_source_path.setter
    def _source_path(self, value: Union[str, PathLike]) -> None:
        self.__source_path = Path(value).as_posix()

    @property
    def id(self) -> Optional[str]:
        """The resource ID.

        :return: The global ID of the resource, an Azure Resource Manager (ARM) ID.
        :rtype: Optional[str]
        """
        if self._id is None:
            return None
        return str(self._id)

    @property
    def creation_context(self) -> Optional[SystemData]:
        """The creation context of the resource.

        :return: The creation metadata for the resource.
        :rtype: Optional[~azure.ai.ml.entities.SystemData]
        """
        return cast(Optional[SystemData], self._creation_context)

    @property
    def base_path(self) -> str:
        """The base path of the resource.

        :return: The base path of the resource.
        :rtype: str
        """
        return self._base_path

    @abc.abstractmethod
    def dump(self, dest: Union[str, PathLike, IO[AnyStr]], **kwargs: Any) -> Any:
        """Dump the object content into a file.

        :param dest: The local path or file stream to write the YAML content to.
            If dest is a file path, a new file will be created.
            If dest is an open file, the file will be written to directly.
        :type dest: Union[PathLike, str, IO[AnyStr]]
        """

    @classmethod
    # pylint: disable=unused-argument
    def _resolve_cls_and_type(cls, data: Dict, params_override: Optional[List[Dict]] = None) -> Tuple:
        """Resolve the class to use for deserializing the data. Return current class if no override is provided.

        :param data: Data to deserialize.
        :type data: dict
        :param params_override: Parameters to override, defaults to None
        :type params_override: typing.Optional[list]
        :return: Class to use for deserializing the data & its "type". Type will be None if no override is provided.
        :rtype: tuple[class, typing.Optional[str]]
        """
        return cls, None

    @classmethod
    @abc.abstractmethod
    def _load(
        cls,
        data: Optional[Dict] = None,
        yaml_path: Optional[Union[PathLike, str]] = None,
        params_override: Optional[list] = None,
        **kwargs: Any,
    ) -> "Resource":
        """Construct a resource object from a file. @classmethod.

        :param cls: Indicates that this is a class method.
        :type cls: class
        :param data: Path to a local file as the source, defaults to None
        :type data: typing.Optional[typing.Dict]
        :param yaml_path: Path to a yaml file as the source, defaults to None
        :type yaml_path: typing.Optional[typing.Union[typing.PathLike, str]]
        :param params_override: Parameters to override, defaults to None
        :type params_override: typing.Optional[list]
        :return: Resource
        :rtype: Resource
        """

    # pylint: disable:unused-argument
    def _get_arm_resource(
        self,
        # pylint: disable=unused-argument
        **kwargs: Any,
    ) -> Dict:
        """Get arm resource.

        :return: Resource
        :rtype: dict
        """
        from azure.ai.ml._arm_deployments.arm_helper import get_template

        # pylint: disable=no-member
        template = get_template(resource_type=self._arm_type)  # type: ignore
        # pylint: disable=no-member
        template["copy"]["name"] = f"{self._arm_type}Deployment"  # type: ignore
        return dict(template)

    def _get_arm_resource_and_params(self, **kwargs: Any) -> List:
        """Get arm resource and parameters.

        :return: Resource and parameters
        :rtype: dict
        """
        resource = self._get_arm_resource(**kwargs)
        # pylint: disable=no-member
        param = self._to_arm_resource_param(**kwargs)  # type: ignore
        return [(resource, param)]

    def __repr__(self) -> str:
        var_dict = {k.strip("_"): v for (k, v) in vars(self).items()}
        return f"{self.__class__.__name__}({var_dict})"

    def __str__(self) -> str:
        if self._print_as_yaml or in_jupyter_notebook():
            # pylint: disable=no-member
            yaml_serialized = self._to_dict()  # type: ignore
            return str(dump_yaml(yaml_serialized, default_flow_style=False))
        return self.__repr__()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import os
from typing import Any, Dict, List, Optional, Union

from marshmallow import Schema

from azure.ai.ml._schema.component.spark_component import SparkComponentSchema
from azure.ai.ml.constants._common import COMPONENT_TYPE
from azure.ai.ml.constants._component import NodeType
from azure.ai.ml.constants._job.job import RestSparkConfKey
from azure.ai.ml.entities._assets import Environment
from azure.ai.ml.entities._job.parameterized_spark import ParameterizedSpark

from ..._schema import PathAwareSchema
from .._job.spark_job_entry_mixin import SparkJobEntry, SparkJobEntryMixin
from .._util import convert_ordered_dict_to_dict, validate_attribute_type
from .._validation import MutableValidationResult
from ._additional_includes import AdditionalIncludesMixin
from .component import Component


class SparkComponent(
    Component, ParameterizedSpark, SparkJobEntryMixin, AdditionalIncludesMixin
):  # pylint: disable=too-many-instance-attributes
    """Spark component version, used to define a Spark Component or Job.

    :keyword code: The source code to run the job. Can be a local path or "http:", "https:", or "azureml:" url pointing
        to a remote location. Defaults to ".", indicating the current directory.
    :type code: Union[str, os.PathLike]
    :keyword entry: The file or class entry point.
    :paramtype entry: Optional[Union[dict[str, str], ~azure.ai.ml.entities.SparkJobEntry]]
    :keyword py_files: The list of .zip, .egg or .py files to place on the PYTHONPATH for Python apps. Defaults to None.
    :paramtype py_files: Optional[List[str]]
    :keyword jars: The list of .JAR files to include on the driver and executor classpaths. Defaults to None.
    :paramtype jars: Optional[List[str]]
    :keyword files: The list of files to be placed in the working directory of each executor. Defaults to None.
    :paramtype files: Optional[List[str]]
    :keyword archives: The list of archives to be extracted into the working directory of each executor.
        Defaults to None.
    :paramtype archives: Optional[List[str]]
    :keyword driver_cores: The number of cores to use for the driver process, only in cluster mode.
    :paramtype driver_cores: Optional[int]
    :keyword driver_memory: The amount of memory to use for the driver process, formatted as strings with a size unit
        suffix ("k", "m", "g" or "t") (e.g. "512m", "2g").
    :paramtype driver_memory: Optional[str]
    :keyword executor_cores: The number of cores to use on each executor.
    :paramtype executor_cores: Optional[int]
    :keyword executor_memory: The amount of memory to use per executor process, formatted as strings with a size unit
        suffix ("k", "m", "g" or "t") (e.g. "512m", "2g").
    :paramtype executor_memory: Optional[str]
    :keyword executor_instances: The initial number of executors.
    :paramtype executor_instances: Optional[int]
    :keyword dynamic_allocation_enabled: Whether to use dynamic resource allocation, which scales the number of
        executors registered with this application up and down based on the workload. Defaults to False.
    :paramtype dynamic_allocation_enabled: Optional[bool]
    :keyword dynamic_allocation_min_executors: The lower bound for the number of executors if dynamic allocation is
        enabled.
    :paramtype dynamic_allocation_min_executors: Optional[int]
    :keyword dynamic_allocation_max_executors: The upper bound for the number of executors if dynamic allocation is
        enabled.
    :paramtype dynamic_allocation_max_executors: Optional[int]
    :keyword conf: A dictionary with pre-defined Spark configurations key and values. Defaults to None.
    :paramtype conf: Optional[dict[str, str]]
    :keyword environment: The Azure ML environment to run the job in.
    :paramtype environment: Optional[Union[str, ~azure.ai.ml.entities.Environment]]
    :keyword inputs: A mapping of input names to input data sources used in the job. Defaults to None.
    :paramtype inputs: Optional[dict[str, Union[
        ~azure.ai.ml.entities._job.pipeline._io.NodeOutput,
        ~azure.ai.ml.Input,
        str,
        bool,
        int,
        float,
        Enum,
        ]]]
    :keyword outputs: A mapping of output names to output data sources used in the job. Defaults to None.
    :paramtype outputs: Optional[dict[str, Union[str, ~azure.ai.ml.Output]]]
    :keyword args: The arguments for the job. Defaults to None.
    :paramtype args: Optional[str]
    :keyword additional_includes: A list of shared additional files to be included in the component. Defaults to None.
    :paramtype additional_includes: Optional[List[str]]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_spark_configurations.py
            :start-after: [START spark_component_definition]
            :end-before: [END spark_component_definition]
            :language: python
            :dedent: 8
            :caption: Creating SparkComponent.
    """

    def __init__(
        self,
        *,
        code: Optional[Union[str, os.PathLike]] = ".",
        entry: Optional[Union[Dict[str, str], SparkJobEntry]] = None,
        py_files: Optional[List[str]] = None,
        jars: Optional[List[str]] = None,
        files: Optional[List[str]] = None,
        archives: Optional[List[str]] = None,
        driver_cores: Optional[Union[int, str]] = None,
        driver_memory: Optional[str] = None,
        executor_cores: Optional[Union[int, str]] = None,
        executor_memory: Optional[str] = None,
        executor_instances: Optional[Union[int, str]] = None,
        dynamic_allocation_enabled: Optional[Union[bool, str]] = None,
        dynamic_allocation_min_executors: Optional[Union[int, str]] = None,
        dynamic_allocation_max_executors: Optional[Union[int, str]] = None,
        conf: Optional[Dict[str, str]] = None,
        environment: Optional[Union[str, Environment]] = None,
        inputs: Optional[Dict] = None,
        outputs: Optional[Dict] = None,
        args: Optional[str] = None,
        additional_includes: Optional[List] = None,
        **kwargs: Any,
    ) -> None:
        # validate init params are valid type
        validate_attribute_type(attrs_to_check=locals(), attr_type_map=self._attr_type_map())

        kwargs[COMPONENT_TYPE] = NodeType.SPARK

        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

        self.code: Optional[Union[str, os.PathLike]] = code
        self.entry = entry
        self.py_files = py_files
        self.jars = jars
        self.files = files
        self.archives = archives
        self.conf = conf
        self.environment = environment
        self.args = args
        self.additional_includes = additional_includes or []
        # For pipeline spark job, we also allow user to set driver_cores, driver_memory and so on by setting conf.
        # If root level fields are not set by user, we promote conf setting to root level to facilitate subsequent
        # verification. This usually happens when we use to_component(SparkJob) or builder function spark() as a node
        # in pipeline sdk
        conf = conf or {}
        self.driver_cores = driver_cores or conf.get(RestSparkConfKey.DRIVER_CORES, None)
        self.driver_memory = driver_memory or conf.get(RestSparkConfKey.DRIVER_MEMORY, None)
        self.executor_cores = executor_cores or conf.get(RestSparkConfKey.EXECUTOR_CORES, None)
        self.executor_memory = executor_memory or conf.get(RestSparkConfKey.EXECUTOR_MEMORY, None)
        self.executor_instances = executor_instances or conf.get(RestSparkConfKey.EXECUTOR_INSTANCES, None)
        self.dynamic_allocation_enabled = dynamic_allocation_enabled or conf.get(
            RestSparkConfKey.DYNAMIC_ALLOCATION_ENABLED, None
        )
        self.dynamic_allocation_min_executors = dynamic_allocation_min_executors or conf.get(
            RestSparkConfKey.DYNAMIC_ALLOCATION_MIN_EXECUTORS, None
        )
        self.dynamic_allocation_max_executors = dynamic_allocation_max_executors or conf.get(
            RestSparkConfKey.DYNAMIC_ALLOCATION_MAX_EXECUTORS, None
        )

    @classmethod
    def _create_schema_for_validation(cls, context: Any) -> Union[PathAwareSchema, Schema]:
        return SparkComponentSchema(context=context)

    @classmethod
    def _attr_type_map(cls) -> dict:
        return {
            "environment": (str, Environment),
            "code": (str, os.PathLike),
        }

    def _customized_validate(self) -> MutableValidationResult:
        validation_result = super()._customized_validate()
        self._append_diagnostics_and_check_if_origin_code_reliable_for_local_path_validation(validation_result)
        return validation_result

    def _to_dict(self) -> Dict:
        # TODO: Bug Item number: 2897665
        res: Dict = convert_ordered_dict_to_dict(  # type: ignore
            {**self._other_parameter, **super(SparkComponent, self)._to_dict()}
        )
        return res

    def _to_ordered_dict_for_yaml_dump(self) -> Dict:
        """Dump the component content into a sorted yaml string.

        :return: The ordered dict
        :rtype: Dict
        """

        obj: dict = super()._to_ordered_dict_for_yaml_dump()
        # dict dumped base on schema will transfer code to an absolute path, while we want to keep its original value
        if self.code and isinstance(self.code, str):
            obj["code"] = self.code
        return obj

    def _get_environment_id(self) -> Union[str, None]:
        # Return environment id of environment
        # handle case when environment is defined inline
        if isinstance(self.environment, Environment):
            res: Optional[str] = self.environment.id
            return res
        return self.environment

    def __str__(self) -> str:
        try:
            toYaml: str = self._to_yaml()
            return toYaml
        except BaseException:  # pylint: disable=W0718
            toStr: str = super(SparkComponent, self).__str__()
            return toStr



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
import random
import uuid

from azure.ai.ml._azure_environments import _get_base_url_from_metadata, _resource_to_scopes
from azure.ai.ml._vendor.azure_resources._resource_management_client import ResourceManagementClient
from azure.ai.ml._vendor.azure_resources.models import GenericResource
from azure.ai.ml.constants._common import ArmConstants
from azure.core.credentials import TokenCredential

module_logger = logging.getLogger(__name__)


def get_name_for_dependent_resource(workspace_name: str, resource_type: str) -> str:
    alphabets_str = ""
    for char in workspace_name.lower():
        if char.isalpha() or char.isdigit():
            alphabets_str = alphabets_str + char
    rand_str = str(uuid.uuid4()).replace("-", "")
    resource_name = alphabets_str[:8] + resource_type[:8] + rand_str
    return resource_name[:24]


def get_deployment_name(name: str):
    random.seed(version=2)
    return f"{name}-{random.randint(1, 10000000)}"


def get_resource_group_location(credentials: TokenCredential, subscription_id: str, resource_group_name: str) -> str:
    arm_hostname = _get_base_url_from_metadata()
    client = ResourceManagementClient(
        credential=credentials,
        subscription_id=subscription_id,
        base_url=arm_hostname,
        api_version=ArmConstants.AZURE_MGMT_RESOURCE_API_VERSION,
        credential_scopes=_resource_to_scopes(arm_hostname),
    )
    rg = client.resource_groups.get(resource_group_name)
    return rg.location


def get_generic_arm_resource_by_arm_id(
    credentials: TokenCredential,
    subscription_id: str,
    arm_id: str,
    api_version: str,
) -> GenericResource:
    if arm_id:
        arm_hostname = _get_base_url_from_metadata()
        client = ResourceManagementClient(
            credential=credentials,
            subscription_id=subscription_id,
            base_url=arm_hostname,
            api_version=ArmConstants.AZURE_MGMT_RESOURCE_API_VERSION,
            credential_scopes=_resource_to_scopes(arm_hostname),
        )
        return client.resources.get_by_id(arm_id, api_version)
    return None


def delete_resource_by_arm_id(
    credentials: TokenCredential,
    subscription_id: str,
    arm_id: str,
    api_version: str,
) -> None:
    if arm_id:
        arm_hostname = _get_base_url_from_metadata()
        client = ResourceManagementClient(
            credential=credentials,
            subscription_id=subscription_id,
            base_url=arm_hostname,
            api_version=ArmConstants.AZURE_MGMT_RESOURCE_API_VERSION,
            credential_scopes=_resource_to_scopes(arm_hostname),
        )
        client.resources.begin_delete_by_id(arm_id, api_version)


def get_resource_and_group_name(armstr: str) -> str:
    return armstr.split("/")[-1], armstr.split("/")[-5]


def get_sub_id_resource_and_group_name(armstr: str) -> str:
    return armstr.split("/")[-7], armstr.split("/")[-1], armstr.split("/")[-5]


def get_endpoint_parts(arm_id: str, subnet_arm_id: str) -> ():
    arm_id_parts = arm_id.split("/")
    subnet_id_parts = subnet_arm_id.split("/")
    conn_name = arm_id_parts[-1]
    subscription_id = arm_id_parts[2]
    resource_group = arm_id_parts[4]
    vnet_name = subnet_id_parts[-3]
    subnet_name = subnet_id_parts[-1]
    return conn_name, subscription_id, resource_group, vnet_name, subnet_name



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import json
from pathlib import Path
from typing import Any, Dict

import pydash
import pytest
import yaml
from azure.ai.ml import Input, load_component, load_job
from azure.ai.ml._internal import (
    Ae365exepool,
    AISuperComputerConfiguration,
    AISuperComputerScalePolicy,
    AISuperComputerStorageReferenceConfiguration,
    Command,
    DataTransfer,
    Distributed,
    HDInsight,
    Hemera,
    ITPConfiguration,
    ITPInteractiveConfiguration,
    ITPPriorityConfiguration,
    ITPResourceConfiguration,
    ITPRetrySettings,
    Parallel,
    Pipeline,
    Scope,
    Starlite,
    TargetSelector,
)
from azure.ai.ml._internal.entities import InternalBaseNode, InternalComponent, Scope
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.constants._job.job import JobComputePropertyFields
from azure.ai.ml.dsl import pipeline
from azure.ai.ml.entities import (
    CommandComponent,
    CommandJobLimits,
    Data,
    JobResourceConfiguration,
    PipelineJob,
    SparkResourceConfiguration,
)
from test_utilities.utils import parse_local_path

from .._utils import (
    DATA_VERSION,
    PARAMETERS_TO_TEST,
    assert_strong_type_intellisense_enabled,
    extract_non_primitive,
    get_expected_runsettings_items,
    set_run_settings,
)


@pytest.mark.usefixtures("enable_internal_components", "enable_pipeline_private_preview_features")
@pytest.mark.unittest
@pytest.mark.pipeline_test
class TestPipelineJob:
    @pytest.mark.parametrize(
        "yaml_path,inputs,runsettings_dict,pipeline_runsettings_dict",
        PARAMETERS_TO_TEST,
    )
    def test_anonymous_internal_component_in_pipeline(
        self, yaml_path, inputs, runsettings_dict, pipeline_runsettings_dict
    ):
        node_func: InternalComponent = load_component(yaml_path)

        @pipeline()
        def pipeline_func():
            node = node_func(**inputs)
            set_run_settings(node, runsettings_dict)
            assert_strong_type_intellisense_enabled(node, runsettings_dict)

        dsl_pipeline: PipelineJob = pipeline_func()
        set_run_settings(dsl_pipeline.settings, pipeline_runsettings_dict)
        result = dsl_pipeline._validate()
        assert result._to_dict() == {"result": "Succeeded"}

        # check if node type is correct
        node_in_pipeline = dsl_pipeline.jobs["node"]
        if node_func.type == "CommandComponent":
            assert isinstance(node_in_pipeline, Command)
        elif node_func.type == "ScopeComponent":
            assert isinstance(node_in_pipeline, Scope)
        elif node_func.type == "HDInsightComponent":
            assert isinstance(node_in_pipeline, HDInsight)
        elif node_func.type == "ParallelComponent":
            assert isinstance(node_in_pipeline, Parallel)
        elif node_func.type == "DistributedComponent":
            assert isinstance(node_in_pipeline, Distributed)
        elif node_func.type == "DataTransferComponent":
            assert isinstance(node_in_pipeline, DataTransfer)
        elif node_func.type == "StarliteComponent":
            assert isinstance(node_in_pipeline, Starlite)
        elif node_func.type == "PipelineComponent":
            assert isinstance(node_in_pipeline, Pipeline)
        elif node_func.type == "HemeraComponent":
            assert isinstance(node_in_pipeline, Hemera)
        elif node_func.type == "Ae365exepoolComponent":
            assert isinstance(node_in_pipeline, Ae365exepool)

        # check if node's runsettings are set correctly
        node_rest_dict = dsl_pipeline._to_rest_object().properties.jobs["node"]
        del node_rest_dict["componentId"]  # delete component spec to make it a pure dict
        mismatched_runsettings = {}
        for dot_key, expected_value in get_expected_runsettings_items(runsettings_dict):
            value = pydash.get(node_rest_dict, dot_key)
            if value != expected_value:
                mismatched_runsettings[dot_key] = (value, expected_value)
        assert not mismatched_runsettings, "Current value:\n{}\nMismatched fields:\n{}".format(
            json.dumps(node_rest_dict, indent=2), json.dumps(mismatched_runsettings, indent=2)
        )

        pipeline_dict = dsl_pipeline._to_dict()
        pipeline_dict = extract_non_primitive(pipeline_dict)
        assert not pipeline_dict, pipeline_dict

    @pytest.mark.parametrize(
        "yaml_path,inputs,runsettings_dict,pipeline_runsettings_dict",
        PARAMETERS_TO_TEST,
    )
    def test_data_as_node_inputs(self, yaml_path, inputs, runsettings_dict, pipeline_runsettings_dict):
        node_func: InternalComponent = load_component(yaml_path)

        input_data_names = {}
        for input_name, input_obj in inputs.items():
            if isinstance(input_obj, Input):
                data_name = input_obj.path.split("@")[0]
                if "spark" in yaml_path:
                    input_data_names[input_name] = input_obj
                else:
                    inputs[input_name] = Data(name=data_name, version=DATA_VERSION, type=AssetTypes.MLTABLE)
                    input_data_names[input_name] = data_name
        if len(input_data_names) == 0:
            return

        @pipeline()
        def pipeline_func():
            node = node_func(**inputs)
            set_run_settings(node, runsettings_dict)
            assert_strong_type_intellisense_enabled(node, runsettings_dict)

        dsl_pipeline: PipelineJob = pipeline_func()
        set_run_settings(dsl_pipeline.settings, pipeline_runsettings_dict)
        assert dsl_pipeline._validate().passed, repr(dsl_pipeline._validate())

        node_rest_dict = dsl_pipeline._to_rest_object().properties.jobs["node"]
        for input_name, dataset_name in input_data_names.items():
            if "spark" in yaml_path:
                expected_rest_obj = {"job_input_type": AssetTypes.MLTABLE, "uri": dataset_name.path, "mode": "Direct"}
            else:
                expected_rest_obj = {
                    "job_input_type": AssetTypes.MLTABLE,
                    "uri": dataset_name + ":" + DATA_VERSION,
                }
            assert node_rest_dict["inputs"][input_name] == expected_rest_obj

    @pytest.mark.parametrize(
        "input_path",
        [
            "test:" + DATA_VERSION,
            "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/rg/providers/"
            "Microsoft.MachineLearningServices/workspaces/ws/datasets/test/versions/" + DATA_VERSION,
        ],
    )
    def test_data_as_pipeline_inputs(self, input_path):
        yaml_path = "./tests/test_configs/internal/distribution-component/component_spec.yaml"
        node_func: InternalComponent = load_component(yaml_path)

        @pipeline()
        def pipeline_func(pipeline_input):
            node = node_func(input_path=pipeline_input)
            node.compute = "cpu-cluster"

        dsl_pipeline: PipelineJob = pipeline_func(pipeline_input=Input(path=input_path, type=AssetTypes.MLTABLE))
        if input_path.startswith("test:"):
            dsl_pipeline_with_data_input: PipelineJob = pipeline_func(
                pipeline_input=Data(name="test", version=DATA_VERSION, type=AssetTypes.MLTABLE)
            )
        else:
            dsl_pipeline_with_data_input: PipelineJob = pipeline_func(
                pipeline_input=Data(name="test", id=input_path, type=AssetTypes.MLTABLE)
            )
        result = dsl_pipeline_with_data_input._validate()
        assert result._to_dict() == {"result": "Succeeded"}

        assert (
            dsl_pipeline_with_data_input._to_rest_object().properties.as_dict()
            == dsl_pipeline._to_rest_object().properties.as_dict()
        )

        pipeline_rest_dict = dsl_pipeline_with_data_input._to_rest_object().properties

        expected_rest_obj = {
            "job_input_type": AssetTypes.MLTABLE,
            "uri": input_path,
        }
        assert pipeline_rest_dict.inputs["pipeline_input"].as_dict() == expected_rest_obj

        expected_rest_obj = {
            "job_input_type": "literal",
            "value": "${{parent.inputs.pipeline_input}}",
        }
        assert pipeline_rest_dict.jobs["node"]["inputs"]["input_path"] == expected_rest_obj

    def test_internal_component_node_output_type(self):
        from azure.ai.ml._utils.utils import try_enable_internal_components

        # force register internal components after partially reload schema files
        try_enable_internal_components(force=True)
        yaml_path = "./tests/test_configs/internal/component_with_input_outputs/component_spec.yaml"
        component_func = load_component(yaml_path)

        @pipeline
        def pipeline_func():
            node = component_func()
            # node level should have correct output type when type not configured
            node.outputs.data_any_file.mode = "mount"

        pipeline_job = pipeline_func()
        rest_pipeline_job_dict = pipeline_job._to_rest_object().as_dict()
        assert rest_pipeline_job_dict["properties"]["jobs"]["node"]["outputs"] == {
            "data_any_file": {"mode": "ReadWriteMount", "job_output_type": "uri_file"}
        }

    def test_internal_component_output_as_pipeline_component_output(self):
        from azure.ai.ml._utils.utils import try_enable_internal_components

        # force register internal components after partially reload schema files
        try_enable_internal_components(force=True)

        yaml_path = "./tests/test_configs/internal/component_with_input_outputs/component_spec.yaml"
        component_func = load_component(yaml_path, params_override=[{"inputs": {}}])

        @pipeline()
        def sub_pipeline_func():
            node = component_func()
            node.adla_account_name = "adla_account_name"
            return node.outputs

        sub_pipeline_1 = sub_pipeline_func()
        assert sub_pipeline_1._validate().passed
        assert sub_pipeline_1.outputs["data_any_file"].type == "uri_file"

        # confirm that the output won't change in further call
        sub_pipeline_2 = sub_pipeline_func()
        assert sub_pipeline_2.outputs["data_any_file"].type == "uri_file"

        @pipeline()
        def pipeline_func():
            sub_pipeline_func()

        dsl_pipeline: PipelineJob = pipeline_func()
        assert dsl_pipeline._validate().passed
        dsl_pipeline._to_rest_object()
        pipeline_component = dsl_pipeline.jobs["sub_pipeline_func"].component
        assert pipeline_component._to_rest_object().properties.component_spec["outputs"] == {
            "data_any_directory": {"type": "uri_folder"},
            "data_any_file": {"type": "uri_file"},  # AnyFile => uri_file
            "data_azureml_dataset": {"type": "uri_folder"},
            "data_cosmos_structured_stream": {"type": "uri_folder"},
            "data_csv_file": {"type": "uri_folder"},
            "data_data_frame_directory": {"type": "uri_folder"},
            "data_image_directory": {"type": "uri_folder"},
            "data_model_directory": {"type": "uri_folder"},
            "data_path": {"type": "uri_folder"},
            "data_transformation_directory": {"type": "uri_folder"},
            "data_untrained_model_directory": {"type": "uri_folder"},
            "data_zip_file": {"type": "uri_folder"},
        }

    def test_ipp_internal_component_in_pipeline(self):
        yaml_path = "./tests/test_configs/internal/ipp-component/spec.yaml"
        # TODO: support anonymous ipp component creation
        # curated env with name & version
        # command_func: InternalComponent = client.components.get("ls_command", "0.0.1")
        node_func: InternalComponent = load_component(yaml_path)

        @pipeline()
        def pipeline_func():
            node_func()

        dsl_pipeline: PipelineJob = pipeline_func()
        result = dsl_pipeline._validate()
        assert result._to_dict() == {"result": "Succeeded"}

    def test_gjd_internal_component_in_pipeline(self):
        yaml_path = "./tests/test_configs/internal/command-component-ls/ls_command_component.yaml"  # GJD is based on CommandComponent
        node_func: CommandComponent = load_component(yaml_path)
        node = node_func()

        ts = TargetSelector(
            compute_type="AMLK8s",  # runsettings.target_selector.compute_type
            instance_types=["STANDARD_D2_V2"],  # runsettings.target_selector.instance_types
            regions=["eastus2euap"],  # runsettings.target_selector.regions
            my_resource_only=True,  # runsettings.target_selector.my_resource_only
            allow_spot_vm=True,  # runsettings.target_selector.allow_spot_vm
        )
        node.resources.properties["target_selector"] = ts

        properties_rest_dict = node._to_rest_object()["resources"]["properties"]
        assert properties_rest_dict == {
            "target_selector": {
                "compute_type": "AMLK8s",
                "instance_types": ["STANDARD_D2_V2"],
                "regions": ["eastus2euap"],
                "my_resource_only": True,
                "allow_spot_vm": True,
            }
        }

    def test_elastic_component_in_pipeline(self):
        yaml_path = "./tests/test_configs/internal/command-component-ls/ls_command_component.yaml"  # itp & elastic are based on CommandComponent
        node_func: CommandComponent = load_component(yaml_path)
        node = node_func()
        configuration = ITPConfiguration(
            resource_configuration=ITPResourceConfiguration(
                gpu_count=2,
                cpu_count=2,
                memory_request_in_gb=2,
            ),
            priority_configuration=ITPPriorityConfiguration(
                job_priority=1,
                is_preemptible=True,
                node_count_set=[1, 2, 3],
                scale_interval=60,
            ),
            interactive_configuration=ITPInteractiveConfiguration(
                is_ssh_enabled=True,
                ssh_public_key="ssh_key",
                is_i_python_enabled=True,
                is_tensor_board_enabled=True,
                interactive_port=40000,
            ),
            retry=ITPRetrySettings(
                max_retry_count=1,
            ),
        )
        node.resources.properties["itp"] = configuration
        properties_rest_dict = node._to_rest_object()["resources"]["properties"]
        assert properties_rest_dict == {
            "itp": {
                "resource_configuration": {
                    "gpu_count": 2,
                    "cpu_count": 2,
                    "memory_request_in_gb": 2,
                },
                "priority_configuration": {
                    "job_priority": 1,
                    "is_preemptible": True,
                    "node_count_set": [1, 2, 3],
                    "scale_interval": 60,
                },
                "interactive_configuration": {
                    "interactive_port": 40000,
                    "is_i_python_enabled": True,
                    "is_ssh_enabled": True,
                    "is_tensor_board_enabled": True,
                    "ssh_public_key": "ssh_key",
                },
                "retry": {
                    "max_retry_count": 1,
                },
            }
        }

    def test_singularity_component_in_pipeline(self):
        yaml_path = "./tests/test_configs/internal/command-component-ls/ls_command_component.yaml"  # singularity is based on CommandComponent
        node_func: CommandComponent = load_component(yaml_path)
        node = node_func()
        configuration = AISuperComputerConfiguration(
            instance_type="STANDARD_D2_V2",
            image_version="1.0",
            locations=["eastus2euap"],
            ai_super_computer_storage_data={
                "data": AISuperComputerStorageReferenceConfiguration(
                    container_name="container_name",
                    relative_path="relative_path",
                )
            },
            interactive=True,
            scale_policy=AISuperComputerScalePolicy(
                auto_scale_instance_type_count_set=[1, 2, 3],
                auto_scale_interval_in_sec=60,
                max_instance_type_count=3,
                min_instance_type_count=1,
            ),
            virtual_cluster_arm_id="virtual_cluster_arm_id",
            tensorboard_log_directory="tensorboard_log_directory",
            ssh_public_key="ssh_public_key",
            enable_azml_int=True,
            priority="Medium",
            sla_tier="Standard",
            suspend_on_idle_time_hours=1,
            user_alias="user_alias",
        )
        # what if key is not in lower case?
        node.resources.properties[JobComputePropertyFields.SINGULARITY.lower()] = configuration
        properties_rest_dict = node._to_rest_object()["resources"]["properties"]
        assert properties_rest_dict == {
            JobComputePropertyFields.AISUPERCOMPUTER: {
                "InstanceType": "STANDARD_D2_V2",
                "ImageVersion": "1.0",
                "Locations": ["eastus2euap"],
                "AISuperComputerStorageData": {
                    "data": {
                        "ContainerName": "container_name",
                        "RelativePath": "relative_path",
                    }
                },
                "Interactive": True,
                "ScalePolicy": {
                    "AutoScaleInstanceTypeCountSet": [1, 2, 3],
                    "AutoScaleIntervalInSec": 60,
                    "MaxInstanceTypeCount": 3,
                    "MinInstanceTypeCount": 1,
                },
                "VirtualClusterArmId": "virtual_cluster_arm_id",
                "TensorboardLogDirectory": "tensorboard_log_directory",
                "SSHPublicKey": "ssh_public_key",
                "EnableAzmlInt": True,
                "Priority": "Medium",
                "SLATier": "Standard",
                "SuspendOnIdleTimeHours": 1,
                "UserAlias": "user_alias",
            }
        }

    def test_load_pipeline_job_with_internal_components_as_node(self):
        yaml_path = Path("./tests/test_configs/internal/helloworld/helloworld_component_scope.yml")
        scope_internal_func = load_component(source=yaml_path)
        with open(yaml_path, encoding="utf-8") as yaml_file:
            yaml_dict = yaml.safe_load(yaml_file)

        yaml_dict["code"] = parse_local_path(yaml_dict["code"], scope_internal_func.base_path)

        command_func = load_component("./tests/test_configs/components/helloworld_component.yml")

        @pipeline()
        def pipeline_func():
            node = command_func(component_in_path=Input(path="./tests/test_configs/data"))
            node.compute = "cpu-cluster"

            node_internal: Scope = scope_internal_func(
                TextData=Input(type=AssetTypes.MLTABLE, path="azureml:scope_tsv:1"),
                ExtractionClause="column1:string, column2:int",
            )

            node_internal.priority = 800
            node_internal.adla_account_name = "adla_account_name"
            node_internal.scope_param = "-tokens 50"
            node_internal.custom_job_name_suffix = "component_sdk_test"
            node_internal.properties["AZURE_ML_PathOnCompute_mock_output"] = "mock_path"

        dsl_pipeline: PipelineJob = pipeline_func()
        internal_node_name = "node_internal"
        assert dsl_pipeline.jobs[internal_node_name]._component._to_dict() == yaml_dict

        scope_node: InternalBaseNode = dsl_pipeline.jobs[internal_node_name]
        # TODO: check why need to set base path manually
        scope_node._set_base_path(yaml_path.parent)
        scope_node_dict = scope_node._to_dict()
        assert scope_node_dict == {
            "priority": 800,
            "adla_account_name": "adla_account_name",
            "custom_job_name_suffix": "component_sdk_test",
            "scope_param": "-tokens 50",
            "component": yaml_dict,
            "type": "ScopeComponent",
            "inputs": {
                "ExtractionClause": "column1:string, column2:int",
                "TextData": {"path": "azureml:scope_tsv:1", "type": "mltable"},
            },
            "properties": {"AZURE_ML_PathOnCompute_mock_output": "mock_path"},
        }
        assert pydash.omit(scope_node._to_rest_object(), "componentId") == {
            "_source": "YAML.COMPONENT",
            "priority": 800,
            "adla_account_name": "adla_account_name",
            "custom_job_name_suffix": "component_sdk_test",
            "scope_param": "-tokens 50",
            "inputs": {
                "ExtractionClause": {"job_input_type": "literal", "value": "column1:string, column2:int"},
                "TextData": {"job_input_type": "mltable", "uri": "azureml:scope_tsv:1"},
            },
            "type": "ScopeComponent",
            "properties": {"AZURE_ML_PathOnCompute_mock_output": "mock_path"},
        }
        scope_node._validate(raise_error=True)

        omit_fields = ["jobs.node.component", "jobs.node_internal.component"]
        assert pydash.omit(dsl_pipeline._to_dict(), *omit_fields) == pydash.omit(
            {
                "display_name": "pipeline_func",
                "jobs": {"node": dsl_pipeline.jobs["node"]._to_dict(), "node_internal": scope_node._to_dict()},
                "type": "pipeline",
            },
            *omit_fields,
        )

        dsl_pipeline._validate(raise_error=True)

        rest_object = dsl_pipeline._to_rest_object()
        assert rest_object.properties.jobs == {
            "node": dsl_pipeline.jobs["node"]._to_rest_object(),
            "node_internal": scope_node._to_rest_object(),
        }

    def test_components_in_registry(self):
        command_func = load_component("./tests/test_configs/components/helloworld_component.yml")
        registry_code_id = "azureml://feeds/testFeed/codes/hello_component/versions/0.0.1"

        @pipeline()
        def pipeline_func():
            node = command_func(component_in_path=Input(path="./tests/test_configs/data"))

            # can't create a component from azureml-components in ci workspace
            # so just use a local test
            node._component.code = registry_code_id
            node.compute = "cpu-cluster"

        dsl_pipeline: PipelineJob = pipeline_func()
        assert dsl_pipeline._validate().passed
        regenerated_pipeline_job = PipelineJob._from_rest_object(dsl_pipeline._to_rest_object())
        assert dsl_pipeline._to_dict() == regenerated_pipeline_job._to_dict()

    def test_components_input_output(self):
        yaml_path = "./tests/test_configs/internal/component_with_input_outputs/component_spec.yaml"
        component: InternalComponent = load_component(yaml_path)

        fake_input = Input(type=AssetTypes.MLTABLE, path="azureml:scope_tsv:1")
        inputs = {
            "data_path": fake_input,
            "data_azureml_dataset": fake_input,
            "data_any_directory": fake_input,
            "data_any_file": fake_input,
            "data_zip_file": fake_input,
            "data_csv_file": fake_input,
            "data_transformation_directory": fake_input,
            "data_untrained_model_directory": fake_input,
            "data_image_directory": fake_input,
            "data_model_directory": fake_input,
            "data_data_frame_directory": fake_input,
            "data_cosmos_structured_stream": fake_input,
            "param_string": "xxx",
            "param_string_cap": "xxx",
            "param_int": 1,
            "param_int_cap": 1,
            "param_float": 1.0,
            "param_float_cap": 1.0,
            "param_bool": True,
            "param_bool_cap": False,
            "param_enum": "minimal",
            "param_enum_cap": "reuse",
        }

        @pipeline()
        def pipeline_func():
            node = component(
                **inputs,
            )
            node.adla_account_name = "adla_account_name"

        pipeline_job = pipeline_func()
        assert pipeline_job._validate().passed, pipeline_job._validate()
        rest_obj = pipeline_job._to_rest_object()
        expected_inputs = {
            "param_bool": {"job_input_type": "literal", "value": "True"},
            "param_bool_cap": {"job_input_type": "literal", "value": "False"},
            "param_enum": {"job_input_type": "literal", "value": "minimal"},
            "param_enum_cap": {"job_input_type": "literal", "value": "reuse"},
            "param_float": {"job_input_type": "literal", "value": "1.0"},
            "param_float_cap": {"job_input_type": "literal", "value": "1.0"},
            "param_int": {"job_input_type": "literal", "value": "1"},
            "param_int_cap": {"job_input_type": "literal", "value": "1"},
            "param_string": {"job_input_type": "literal", "value": "xxx"},
            "param_string_cap": {"job_input_type": "literal", "value": "xxx"},
        }
        for key in inputs:
            if key.startswith("data_"):
                expected_inputs[key] = {"job_input_type": "mltable", "uri": "azureml:scope_tsv:1"}
        assert rest_obj.properties.jobs["node"]["inputs"] == expected_inputs

    def test_data_binding_on_node_runsettings(self):
        test_path = "./tests/test_configs/internal/helloworld/helloworld_component_command.yml"
        component: InternalComponent = load_component(test_path)

        @pipeline()
        def pipeline_func(
            compute_name: str = "cpu-cluster", environment_name: str = "AzureML-ACPT-pytorch-1.12-py39-cuda11.6-gpu:8"
        ):
            node = component(
                training_data=Input(path="./tests/test_configs/data"),
                max_epochs=1,
            )
            node.compute = compute_name
            node.environment = environment_name

        pipeline_job = pipeline_func()
        assert pipeline_job._validate().passed, repr(pipeline_job._validate())
        rest_object = pipeline_job._to_rest_object().properties.jobs["node"]
        assert str(rest_object["computeId"]) == "${{parent.inputs.compute_name}}"
        assert str(rest_object["environment"]) == "${{parent.inputs.environment_name}}"

    def test_pipeline_with_setting_node_output_directly(self) -> None:
        component_dir = Path(__file__).parent.parent.parent / "test_configs" / "internal" / "command-component"
        copy_func = load_component(component_dir / "command-linux/copy/component.yaml")

        copy_file = copy_func(
            input_dir=None,
            file_names=None,
        )

        copy_file.outputs.output_dir.path = "path_on_datastore"
        assert copy_file.outputs.output_dir.path == "path_on_datastore"
        assert copy_file.outputs.output_dir.type == "path"

    def test_job_properties(self):
        pipeline_job: PipelineJob = load_job(
            source="./tests/test_configs/internal/pipeline_jobs/pipeline_job_with_properties.yml"
        )
        pipeline_dict = pipeline_job._to_dict()
        rest_pipeline_dict = pipeline_job._to_rest_object().as_dict()["properties"]
        assert pipeline_dict["properties"] == {"AZURE_ML_PathOnCompute_input_data": "/tmp/test"}
        assert rest_pipeline_dict["properties"] == pipeline_dict["properties"]
        for name, node_dict in pipeline_dict["jobs"].items():
            rest_node_dict = rest_pipeline_dict["jobs"][name]
            assert len(node_dict["properties"]) == 1
            assert "AZURE_ML_PathOnCompute_" in list(node_dict["properties"].keys())[0]
            assert node_dict["properties"] == rest_node_dict["properties"]

    @pytest.mark.parametrize(
        "component_path, fields_to_test, fake_inputs",
        [
            pytest.param(
                "./tests/test_configs/internal/command-component-ls/ls_command_component.yaml",
                {
                    "resources.instance_count": JobResourceConfiguration(instance_count=1),
                    "limits.timeout": CommandJobLimits(timeout=100),
                },
                {},
                id="command",
            ),
            pytest.param(
                "./tests/test_configs/internal/batch_inference/batch_score.yaml",
                {
                    "resources.instance_count": JobResourceConfiguration(instance_count=1),
                    "limits.timeout": CommandJobLimits(timeout=100),
                },
                {
                    "model_path": Input(type=AssetTypes.MLTABLE, path="mltable_mnist_model@latest"),
                    "images_to_score": Input(type=AssetTypes.MLTABLE, path="mltable_mnist@latest"),
                },
                id="parallel.resources",
            ),
            pytest.param(
                "tests/test_configs/internal/spark-component/spec.yaml",
                {
                    "resources.runtime_version": SparkResourceConfiguration(runtime_version="2.4"),
                },
                {
                    "file_input1": Input(type=AssetTypes.MLTABLE, path="mltable_mnist@latest", mode="direct"),
                    "file_input2": Input(type=AssetTypes.MLTABLE, path="mltable_mnist@latest", mode="direct"),
                },
                id="spark",
            ),
        ],
    )
    def test_data_binding_expression_on_node_runsettings(
        self, component_path: str, fields_to_test: Dict[str, Any], fake_inputs: Dict[str, Input]
    ):
        component = load_component(component_path)

        @pipeline()
        def pipeline_func(param: str = "2"):
            node = component(**fake_inputs)
            for field, value in fields_to_test.items():
                attr, sub_attr = field.split(".")
                setattr(node, attr, value)
                setattr(getattr(node, attr), sub_attr, str(param))

        pipeline_job: PipelineJob = pipeline_func()
        rest_object = pipeline_job._to_rest_object()
        regenerated_job = PipelineJob._from_rest_object(rest_object)
        expected_dict, actual_dict = pipeline_job._to_dict(), regenerated_job._to_dict()

        assert actual_dict == expected_dict

        # directly update component to arm id
        for _node in pipeline_job.jobs.values():
            _node._component = (
                "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/rg/providers/"
                "Microsoft.MachineLearningServices/workspaces/ws/components/component_name/"
                "versions/1.0.0"
            )
        # check if all the fields are correctly serialized
        pipeline_job.component._get_anonymous_hash()

    def test_node_output_type_promotion(self):
        test_path = "./tests/test_configs/internal/helloworld/helloworld_component_command.yml"
        component: InternalComponent = load_component(test_path)

        @pipeline()
        def pipeline_func():
            node = component(
                training_data=Input(path="./tests/test_configs/data"),
                max_epochs=1,
            )
            node.outputs.model_output.mode = "mount"
            return node.outputs

        pipeline_job = pipeline_func()
        pipeline_dict = pipeline_job._to_rest_object().as_dict()
        # type will be preserved & mode will be promoted to pipeline level
        assert pipeline_dict["properties"]["outputs"]["model_output"] == {
            "job_output_type": "uri_folder",
            "mode": "ReadWriteMount",
        }



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import fields, post_load, pre_load

from azure.ai.ml._schema._utils.utils import replace_key_in_odict
from azure.ai.ml._schema.core.schema import PatchedSchemaMeta

module_logger = logging.getLogger(__name__)


class ResourceSettingsSchema(metaclass=PatchedSchemaMeta):
    cpu = fields.String()
    memory = fields.String()
    gpu = fields.String()

    @pre_load
    def conversion(self, data: Any, **kwargs) -> Any:
        data = replace_key_in_odict(data, "nvidia.com/gpu", "gpu")
        return data

    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:
        from azure.ai.ml.entities import ResourceSettings

        return ResourceSettings(**data)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

# pylint: disable=protected-access

"""Customize generated code here.

Follow our quickstart for examples: https://aka.ms/azsdk/python/dpcodegen/python/customize
"""
import json
from typing import Any, Dict, List, Optional

from azure.ai.ml._restclient.v2024_01_01_preview.models import MarketplaceSubscription as RestMarketplaceSubscription
from azure.ai.ml._restclient.v2024_01_01_preview.models import (
    MarketplaceSubscriptionProperties as RestMarketplaceSubscriptionProperties,
)
from azure.ai.ml._restclient.v2024_01_01_preview.models import ModelSettings as RestModelSettings
from azure.ai.ml._restclient.v2024_01_01_preview.models import ServerlessEndpoint as RestServerlessEndpoint
from azure.ai.ml._restclient.v2024_01_01_preview.models import (
    ServerlessEndpointProperties as RestServerlessEndpointProperties,
)
from azure.ai.ml._restclient.v2024_01_01_preview.models import Sku as RestSku
from azure.ai.ml._restclient.v2024_04_01_preview.models import (
    EndpointDeploymentResourcePropertiesBasicResource,
    OpenAIEndpointDeploymentResourceProperties,
)
from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml._utils.utils import camel_to_snake
from azure.ai.ml.entities._system_data import SystemData

from .._model_base import rest_field
from ._models import AzureOpenAIDeployment as _AzureOpenAIDeployment
from ._models import MarketplacePlan as _MarketplacePlan
from ._models import MarketplaceSubscription as _MarketplaceSubscription
from ._models import ServerlessEndpoint as _ServerlessEndpoint

__all__: List[str] = [
    "AzureOpenAIDeployment",
    "ServerlessEndpoint",
    "MarketplaceSubscription",
    "MarketplacePlan",
]  # Add all objects you want publicly available to users at this package level

_NULL = object()


func_to_attr_type = {
    "_deserialize_dict": dict,
    "_deserialize_sequence": list,
}


def _get_rest_field_type(field):
    if hasattr(field, "_type"):
        if field._type.func.__name__ == "_deserialize_default":
            return field._type.args[0]
        if func_to_attr_type.get(field._type.func.__name__):
            return func_to_attr_type[field._type.func.__name__]
        return _get_rest_field_type(field._type.args[0])
    if hasattr(field, "func") and func_to_attr_type.get(field.func.__name__):
        return func_to_attr_type[field.func.__name__]
    if hasattr(field, "args"):
        return _get_rest_field_type(field.args[0])
    return field


class ValidationMixin:
    def _validate(self) -> None:
        # verify types
        for attr, field in self._attr_to_rest_field.items():  # type: ignore
            try:
                attr_value = self.__getitem__(attr)  # type: ignore
                attr_type = type(attr_value)
            except KeyError as exc:
                if field._visibility and "read" in field._visibility:
                    # read-only field, no need to validate
                    continue
                if field._type.func.__name__ != "_deserialize_with_optional":
                    # i'm required
                    raise ValueError(f"attr {attr} is a required property for {self.__class__.__name__}") from exc
            else:
                if getattr(attr_value, "_is_model", False):
                    attr_value._validate()
                rest_field_type = _get_rest_field_type(field)
                if attr_type != rest_field_type:
                    raise ValueError(f"Type of attr {attr} is of type {attr_type}, not {rest_field_type}")


@experimental
class AzureOpenAIDeployment(_AzureOpenAIDeployment):

    system_data: Optional[SystemData] = rest_field(visibility=["read"])
    """System data of the deployment."""

    @classmethod
    def _from_rest_object(cls, obj: EndpointDeploymentResourcePropertiesBasicResource) -> "AzureOpenAIDeployment":
        properties: OpenAIEndpointDeploymentResourceProperties = obj.properties
        return cls(
            name=obj.name,
            model_name=properties.model.name,
            model_version=properties.model.version,
            id=obj.id,
            system_data=SystemData._from_rest_object(obj.system_data),
        )

    def as_dict(self, *, exclude_readonly: bool = False) -> Dict[str, Any]:
        d = super().as_dict(exclude_readonly=exclude_readonly)
        d["system_data"] = json.loads(json.dumps(self.system_data._to_dict()))  # type: ignore
        return d


AzureOpenAIDeployment.__doc__ += (
    _AzureOpenAIDeployment.__doc__.strip()  # type: ignore
    + """
    :ivar system_data: System data of the deployment.
    :vartype system_data: ~azure.ai.ml.entities.SystemData
"""
)


@experimental
class MarketplacePlan(_MarketplacePlan):
    pass


@experimental
class ServerlessEndpoint(_ServerlessEndpoint, ValidationMixin):

    system_data: Optional[SystemData] = rest_field(visibility=["read"])
    """System data of the endpoint."""

    def _to_rest_object(self) -> RestServerlessEndpoint:
        return RestServerlessEndpoint(
            properties=RestServerlessEndpointProperties(
                model_settings=RestModelSettings(model_id=self.model_id),
            ),
            auth_mode="key",  # only key is supported for now
            tags=self.tags,
            sku=RestSku(name="Consumption"),
            location=self.location,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestServerlessEndpoint) -> "ServerlessEndpoint":
        return cls(  # type: ignore
            name=obj.name,
            id=obj.id,
            tags=obj.tags,
            location=obj.location,
            auth_mode=obj.properties.auth_mode,
            provisioning_state=camel_to_snake(obj.properties.provisioning_state),
            model_id=obj.properties.model_settings.model_id if obj.properties.model_settings else None,
            scoring_uri=obj.properties.inference_endpoint.uri if obj.properties.inference_endpoint else None,
            system_data=SystemData._from_rest_object(obj.system_data) if obj.system_data else None,
            headers=obj.properties.inference_endpoint.headers if obj.properties.inference_endpoint else None,
        )

    def as_dict(self, *, exclude_readonly: bool = False) -> Dict[str, Any]:
        d = super().as_dict(exclude_readonly=exclude_readonly)
        d["system_data"] = json.loads(json.dumps(self.system_data._to_dict()))  # type: ignore
        return d


ServerlessEndpoint.__doc__ += (
    _ServerlessEndpoint.__doc__.strip()  # type: ignore
    + """
    :ivar system_data: System data of the endpoint.
    :vartype system_data: ~azure.ai.ml.entities.SystemData
"""
)


@experimental
class MarketplaceSubscription(_MarketplaceSubscription, ValidationMixin):

    system_data: Optional[SystemData] = rest_field(visibility=["read"])
    """System data of the endpoint."""

    def _to_rest_object(self) -> RestMarketplaceSubscription:
        return RestMarketplaceSubscription(properties=RestMarketplaceSubscriptionProperties(model_id=self.model_id))

    @classmethod
    def _from_rest_object(cls, obj: RestMarketplaceSubscription) -> "MarketplaceSubscription":
        properties = obj.properties
        return cls(  # type: ignore
            name=obj.name,
            id=obj.id,
            model_id=properties.model_id,
            marketplace_plan=MarketplacePlan(
                publisher_id=properties.marketplace_plan.publisher_id,
                offer_id=properties.marketplace_plan.offer_id,
                plan_id=properties.marketplace_plan.plan_id,
            ),
            status=camel_to_snake(properties.marketplace_subscription_status),
            provisioning_state=camel_to_snake(properties.provisioning_state),
            system_data=SystemData._from_rest_object(obj.system_data) if obj.system_data else None,
        )

    def as_dict(self, *, exclude_readonly: bool = False) -> Dict[str, Any]:
        d = super().as_dict(exclude_readonly=exclude_readonly)
        if self.system_data:
            d["system_data"] = json.loads(json.dumps(self.system_data._to_dict()))
        return d


MarketplaceSubscription.__doc__ = (
    _MarketplaceSubscription.__doc__.strip()  # type: ignore
    + """
    :ivar system_data: System data of the marketplace subscription.
    :vartype system_data: ~azure.ai.ml.entities.SystemData
"""
)


def patch_sdk():
    """Do not remove from this file.

    `patch_sdk` is a last resort escape hatch that allows you to do customizations
    you can't accomplish using the techniques described in
    https://aka.ms/azsdk/python/dpcodegen/python/customize
    """



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access, too-many-boolean-expressions

import re
from typing import Any, Optional, TypeVar, Union

from azure.ai.ml._restclient.v2024_01_01_preview import AzureMachineLearningWorkspaces as ServiceClient012024Preview
from azure.ai.ml._scope_dependent_operations import (
    OperationConfig,
    OperationsContainer,
    OperationScope,
    _ScopeDependentOperations,
)
from azure.ai.ml._telemetry import ActivityType, monitor_with_activity
from azure.ai.ml._utils._arm_id_utils import AMLVersionedArmId
from azure.ai.ml._utils._azureml_polling import AzureMLPolling
from azure.ai.ml._utils._endpoint_utils import upload_dependencies, validate_scoring_script
from azure.ai.ml._utils._http_utils import HttpPipeline
from azure.ai.ml._utils._logger_utils import OpsLogger
from azure.ai.ml._utils._package_utils import package_deployment
from azure.ai.ml._utils.utils import _get_mfe_base_url_from_discovery_service, modified_operation_client
from azure.ai.ml.constants._common import ARM_ID_PREFIX, AzureMLResourceType, LROConfigurations
from azure.ai.ml.entities import BatchDeployment, BatchJob, ModelBatchDeployment, PipelineComponent, PipelineJob
from azure.ai.ml.entities._deployment.pipeline_component_batch_deployment import PipelineComponentBatchDeployment
from azure.core.credentials import TokenCredential
from azure.core.exceptions import HttpResponseError, ResourceNotFoundError
from azure.core.paging import ItemPaged
from azure.core.polling import LROPoller
from azure.core.tracing.decorator import distributed_trace

from ._operation_orchestrator import OperationOrchestrator

ops_logger = OpsLogger(__name__)
module_logger = ops_logger.module_logger
DeploymentType = TypeVar(
    "DeploymentType", bound=Union[BatchDeployment, PipelineComponentBatchDeployment, ModelBatchDeployment]
)


class BatchDeploymentOperations(_ScopeDependentOperations):
    """BatchDeploymentOperations.

    You should not instantiate this class directly. Instead, you should create an MLClient instance that instantiates it
    for you and attaches it as an attribute.

    :param operation_scope: Scope variables for the operations classes of an MLClient object.
    :type operation_scope: ~azure.ai.ml._scope_dependent_operations.OperationScope
    :param operation_config: Common configuration for operations classes of an MLClient object.
    :type operation_config: ~azure.ai.ml._scope_dependent_operations.OperationConfig
    :param service_client_05_2022: Service client to allow end users to operate on Azure Machine Learning Workspace
        resources.
    :type service_client_05_2022: ~azure.ai.ml._restclient.v2022_05_01._azure_machine_learning_workspaces.
        AzureMachineLearningWorkspaces
    :param all_operations: All operations classes of an MLClient object.
    :type all_operations: ~azure.ai.ml._scope_dependent_operations.OperationsContainer
    :param credentials: Credential to use for authentication.
    :type credentials: ~azure.core.credentials.TokenCredential
    """

    def __init__(
        self,
        operation_scope: OperationScope,
        operation_config: OperationConfig,
        service_client_01_2024_preview: ServiceClient012024Preview,
        all_operations: OperationsContainer,
        credentials: Optional[TokenCredential] = None,
        **kwargs: Any,
    ):
        super(BatchDeploymentOperations, self).__init__(operation_scope, operation_config)
        ops_logger.update_filter()
        self._batch_deployment = service_client_01_2024_preview.batch_deployments
        self._batch_job_deployment = kwargs.pop("service_client_09_2020_dataplanepreview").batch_job_deployment
        service_client_02_2023_preview = kwargs.pop("service_client_02_2023_preview")
        self._component_batch_deployment_operations = service_client_02_2023_preview.batch_deployments
        self._batch_endpoint_operations = service_client_01_2024_preview.batch_endpoints
        self._component_operations = service_client_02_2023_preview.component_versions
        self._all_operations = all_operations
        self._credentials = credentials
        self._init_kwargs = kwargs

        self._requests_pipeline: HttpPipeline = kwargs.pop("requests_pipeline")

    @distributed_trace
    @monitor_with_activity(ops_logger, "BatchDeployment.BeginCreateOrUpdate", ActivityType.PUBLICAPI)
    def begin_create_or_update(
        self,
        deployment: DeploymentType,
        *,
        skip_script_validation: bool = False,
        **kwargs: Any,
    ) -> LROPoller[DeploymentType]:
        """Create or update a batch deployment.

        :param deployment: The deployment entity.
        :type deployment: ~azure.ai.ml.entities.BatchDeployment
        :keyword skip_script_validation: If set to True, the script validation will be skipped. Defaults to False.
        :paramtype skip_script_validation: bool
        :raises ~azure.ai.ml.exceptions.ValidationException: Raised if BatchDeployment cannot be
            successfully validated. Details will be provided in the error message.
        :raises ~azure.ai.ml.exceptions.AssetException: Raised if BatchDeployment assets
            (e.g. Data, Code, Model, Environment) cannot be successfully validated.
            Details will be provided in the error message.
        :raises ~azure.ai.ml.exceptions.ModelException: Raised if BatchDeployment model
            cannot be successfully validated. Details will be provided in the error message.
        :return: A poller to track the operation status.
        :rtype: ~azure.core.polling.LROPoller[~azure.ai.ml.entities.BatchDeployment]

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_misc.py
                :start-after: [START batch_deployment_operations_begin_create_or_update]
                :end-before: [END batch_deployment_operations_begin_create_or_update]
                :language: python
                :dedent: 8
                :caption: Create example.
        """
        if (
            not skip_script_validation
            and not isinstance(deployment, PipelineComponentBatchDeployment)
            and deployment
            and deployment.code_configuration  # type: ignore
            and not deployment.code_configuration.code.startswith(ARM_ID_PREFIX)  # type: ignore
            and not re.match(AMLVersionedArmId.REGEX_PATTERN, deployment.code_configuration.code)  # type: ignore
        ):
            validate_scoring_script(deployment)
        module_logger.debug("Checking endpoint %s exists", deployment.endpoint_name)
        self._batch_endpoint_operations.get(
            endpoint_name=deployment.endpoint_name,
            resource_group_name=self._resource_group_name,
            workspace_name=self._workspace_name,
        )
        orchestrators = OperationOrchestrator(
            operation_container=self._all_operations,
            operation_scope=self._operation_scope,
            operation_config=self._operation_config,
        )
        if isinstance(deployment, PipelineComponentBatchDeployment):
            self._validate_component(deployment, orchestrators)  # type: ignore
        else:
            upload_dependencies(deployment, orchestrators)
        try:
            location = self._get_workspace_location()
            if kwargs.pop("package_model", False):
                deployment = package_deployment(deployment, self._all_operations.all_operations)
                module_logger.info("\nStarting deployment")
            deployment_rest = deployment._to_rest_object(location=location)
            if isinstance(deployment, PipelineComponentBatchDeployment):  # pylint: disable=no-else-return
                return self._component_batch_deployment_operations.begin_create_or_update(
                    resource_group_name=self._resource_group_name,
                    workspace_name=self._workspace_name,
                    endpoint_name=deployment.endpoint_name,
                    deployment_name=deployment.name,
                    body=deployment_rest,
                    **self._init_kwargs,
                    cls=lambda response, deserialized, headers: PipelineComponentBatchDeployment._from_rest_object(
                        deserialized
                    ),
                )
            else:
                return self._batch_deployment.begin_create_or_update(
                    resource_group_name=self._resource_group_name,
                    workspace_name=self._workspace_name,
                    endpoint_name=deployment.endpoint_name,
                    deployment_name=deployment.name,
                    body=deployment_rest,
                    **self._init_kwargs,
                    cls=lambda response, deserialized, headers: BatchDeployment._from_rest_object(deserialized),
                )
        except Exception as ex:
            raise ex

    @distributed_trace
    @monitor_with_activity(ops_logger, "BatchDeployment.Get", ActivityType.PUBLICAPI)
    def get(self, name: str, endpoint_name: str) -> BatchDeployment:
        """Get a deployment resource.

        :param name: The name of the deployment
        :type name: str
        :param endpoint_name: The name of the endpoint
        :type endpoint_name: str
        :return: A deployment entity
        :rtype: ~azure.ai.ml.entities.BatchDeployment

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_misc.py
                :start-after: [START batch_deployment_operations_get]
                :end-before: [END batch_deployment_operations_get]
                :language: python
                :dedent: 8
                :caption: Get example.
        """
        deployment = BatchDeployment._from_rest_object(
            self._batch_deployment.get(
                endpoint_name=endpoint_name,
                deployment_name=name,
                resource_group_name=self._resource_group_name,
                workspace_name=self._workspace_name,
                **self._init_kwargs,
            )
        )
        deployment.endpoint_name = endpoint_name
        return deployment

    @distributed_trace
    @monitor_with_activity(ops_logger, "BatchDeployment.BeginDelete", ActivityType.PUBLICAPI)
    def begin_delete(self, name: str, endpoint_name: str) -> LROPoller[None]:
        """Delete a batch deployment.

        :param name: Name of the batch deployment.
        :type name: str
        :param endpoint_name: Name of the batch endpoint
        :type endpoint_name: str
        :return: A poller to track the operation status.
        :rtype: ~azure.core.polling.LROPoller[None]

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_misc.py
                :start-after: [START batch_deployment_operations_delete]
                :end-before: [END batch_deployment_operations_delete]
                :language: python
                :dedent: 8
                :caption: Delete example.
        """
        path_format_arguments = {
            "endpointName": name,
            "resourceGroupName": self._resource_group_name,
            "workspaceName": self._workspace_name,
        }

        delete_poller = self._batch_deployment.begin_delete(
            resource_group_name=self._resource_group_name,
            workspace_name=self._workspace_name,
            endpoint_name=endpoint_name,
            deployment_name=name,
            polling=AzureMLPolling(
                LROConfigurations.POLL_INTERVAL,
                path_format_arguments=path_format_arguments,
                **self._init_kwargs,
            ),
            polling_interval=LROConfigurations.POLL_INTERVAL,
            **self._init_kwargs,
        )
        return delete_poller

    @distributed_trace
    @monitor_with_activity(ops_logger, "BatchDeployment.List", ActivityType.PUBLICAPI)
    def list(self, endpoint_name: str) -> ItemPaged[BatchDeployment]:
        """List a deployment resource.

        :param endpoint_name: The name of the endpoint
        :type endpoint_name: str
        :return: An iterator of deployment entities
        :rtype: ~azure.core.paging.ItemPaged[~azure.ai.ml.entities.BatchDeployment]

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_misc.py
                :start-after: [START batch_deployment_operations_list]
                :end-before: [END batch_deployment_operations_list]
                :language: python
                :dedent: 8
                :caption: List deployment resource example.
        """
        return self._batch_deployment.list(
            endpoint_name=endpoint_name,
            resource_group_name=self._resource_group_name,
            workspace_name=self._workspace_name,
            cls=lambda objs: [BatchDeployment._from_rest_object(obj) for obj in objs],
            **self._init_kwargs,
        )

    @distributed_trace
    @monitor_with_activity(ops_logger, "BatchDeployment.ListJobs", ActivityType.PUBLICAPI)
    def list_jobs(self, endpoint_name: str, *, name: Optional[str] = None) -> ItemPaged[BatchJob]:
        """List jobs under the provided batch endpoint deployment. This is only valid for batch endpoint.

        :param endpoint_name: Name of endpoint.
        :type endpoint_name: str
        :keyword name: (Optional) Name of deployment.
        :paramtype name: str
        :raise: Exception if endpoint_type is not BATCH_ENDPOINT_TYPE
        :return: List of jobs
        :rtype: ~azure.core.paging.ItemPaged[~azure.ai.ml.entities.BatchJob]

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_misc.py
                :start-after: [START batch_deployment_operations_list_jobs]
                :end-before: [END batch_deployment_operations_list_jobs]
                :language: python
                :dedent: 8
                :caption: List jobs example.
        """

        workspace_operations = self._all_operations.all_operations[AzureMLResourceType.WORKSPACE]
        mfe_base_uri = _get_mfe_base_url_from_discovery_service(
            workspace_operations, self._workspace_name, self._requests_pipeline
        )

        with modified_operation_client(self._batch_job_deployment, mfe_base_uri):
            result = self._batch_job_deployment.list(
                endpoint_name=endpoint_name,
                deployment_name=name,
                resource_group_name=self._resource_group_name,
                workspace_name=self._workspace_name,
                **self._init_kwargs,
            )

            # This is necessary as the paged result need to be resolved inside the context manager
            return list(result)

    def _get_workspace_location(self) -> str:
        """Get the workspace location

        TODO[TASK 1260265]: can we cache this information and only refresh when the operation_scope is changed?

        :return: The workspace location
        :rtype: str
        """
        return str(
            self._all_operations.all_operations[AzureMLResourceType.WORKSPACE].get(self._workspace_name).location
        )

    def _validate_component(self, deployment: Any, orchestrators: OperationOrchestrator) -> None:
        """Validates that the value provided is associated to an existing component or otherwise we will try to create
        an anonymous component that will be use for batch deployment.

        :param deployment: Batch deployment
        :type deployment: ~azure.ai.ml.entities._deployment.deployment.Deployment
        :param orchestrators: Operation Orchestrator
        :type orchestrators: _operation_orchestrator.OperationOrchestrator
        """
        if isinstance(deployment.component, PipelineComponent):
            try:
                registered_component = self._all_operations.all_operations[AzureMLResourceType.COMPONENT].get(
                    name=deployment.component.name, version=deployment.component.version
                )
                deployment.component = registered_component.id
            except Exception as err:  # pylint: disable=W0718
                if isinstance(err, (ResourceNotFoundError, HttpResponseError)):
                    deployment.component = self._all_operations.all_operations[
                        AzureMLResourceType.COMPONENT
                    ].create_or_update(
                        name=deployment.component.name,
                        resource_group_name=self._resource_group_name,
                        workspace_name=self._workspace_name,
                        component=deployment.component,
                        version=deployment.component.version,
                        **self._init_kwargs,
                    )
                else:
                    raise err
        elif isinstance(deployment.component, str):
            component_id = orchestrators.get_asset_arm_id(
                deployment.component, azureml_type=AzureMLResourceType.COMPONENT
            )
            deployment.component = component_id
        elif isinstance(deployment.job_definition, str):
            job_component = PipelineComponent(source_job_id=deployment.job_definition)
            job_component = self._component_operations.create_or_update(
                name=job_component.name,
                resource_group_name=self._resource_group_name,
                workspace_name=self._workspace_name,
                body=job_component._to_rest_object(),
                version=job_component.version,
                **self._init_kwargs,
            )
            deployment.component = job_component.id

        elif isinstance(deployment.job_definition, PipelineJob):
            try:
                registered_job = self._all_operations.all_operations[AzureMLResourceType.JOB].get(
                    name=deployment.job_definition.name
                )
                if registered_job:
                    job_component = PipelineComponent(source_job_id=registered_job.name)
                    job_component = self._component_operations.create_or_update(
                        name=job_component.name,
                        resource_group_name=self._resource_group_name,
                        workspace_name=self._workspace_name,
                        body=job_component._to_rest_object(),
                        version=job_component.version,
                        **self._init_kwargs,
                    )
                    deployment.component = job_component.id
            except ResourceNotFoundError as err:
                raise err



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import INCLUDE, post_dump, post_load

from azure.ai.ml._schema.job_resource_configuration import JobResourceConfigurationSchema


class ComponentResourceSchema(JobResourceConfigurationSchema):
    class Meta:
        unknown = INCLUDE

    @post_load
    def make(self, data, **kwargs):
        return data

    @post_dump(pass_original=True)
    def dump_override(self, data, original, **kwargs):
        return original



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable

import pytest
from devtools_testutils import AzureRecordedTestCase
from pytest_mock import MockFixture
from test_utilities.utils import assert_job_cancel, wait_until_done

from azure.ai.ml import MLClient, Output, dsl, load_component, load_job
from azure.ai.ml.constants import JobType
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._builders.import_node import Import
from azure.ai.ml.entities._job.import_job import DatabaseImportSource, ImportJob
from azure.ai.ml.entities._job.pipeline.pipeline_job import PipelineJob
from azure.ai.ml.operations._run_history_constants import JobStatus, RunHistoryConstants


@pytest.fixture(autouse=True)
def import_job_enabled(mocker: MockFixture):
    mocker.patch("azure.ai.ml.entities._job.import_job.is_private_preview_enabled", return_value=True)
    mocker.patch("azure.ai.ml.entities._job.pipeline.pipeline_job.is_private_preview_enabled", return_value=True)


@pytest.mark.timeout(600)
@pytest.mark.usefixtures(
    "recorded_test",
    "mock_asset_name",
    "mock_code_hash",
    "mock_component_hash",
    "enable_environment_id_arm_expansion",
)
class TestImportJob(AzureRecordedTestCase):
    @pytest.mark.e2etest
    def test_import_job_submit_cancel(self, client: MLClient) -> None:
        # TODO: need to create a workspace under a e2e-testing-only subscription and resource group

        job = load_job("./tests/test_configs/import_job/import_job_test.yml")
        self.validate_import_job_submit_cancel(job, client)

    @pytest.mark.e2etest
    def test_import_node_submit_cancel(self, client: MLClient) -> None:
        from azure.ai.ml.entities._builders.import_func import import_job

        source = DatabaseImportSource(
            type="azuresqldb", connection="azureml:my_username_password", query="select * from REGION"
        )
        output = Output(type="uri_folder", path="azureml://datastores/workspaceblobstore/paths/output_dir/")
        job = import_job(source=source, output=output)
        self.validate_import_job_submit_cancel(job, client)

    @classmethod
    def validate_import_job_submit_cancel(cls, job: ImportJob, client: MLClient) -> None:
        # TODO: need to create a workspace under a e2e-testing-only subscription and resource group

        def check_before_cancel(_import_job: ImportJob) -> True:
            assert _import_job.status in RunHistoryConstants.IN_PROGRESS_STATUSES

            import_job_2 = client.jobs.get(_import_job.name)
            assert isinstance(import_job_2, ImportJob)
            assert import_job_2.type == _import_job.type
            assert import_job_2.name == _import_job.name
            assert import_job_2.compute == "DataFactory"
            assert import_job_2.source.type == _import_job.source.type
            assert import_job_2.source.connection == _import_job.source.connection
            assert import_job_2.source.query == _import_job.source.query
            assert import_job_2.output.type == _import_job.output.type
            assert import_job_2.output.path == _import_job.output.path
            assert import_job_2.status in [JobStatus.COMPLETED, JobStatus.FAILED, JobStatus.RUNNING, JobStatus.STARTING]
            return True

        import_job = assert_job_cancel(job, client, check_before_cancelled=check_before_cancel)

        assert import_job.type == JobType.IMPORT
        assert import_job.source.type == "azuresqldb"
        assert import_job.source.connection == "azureml:my_username_password"
        assert import_job.source.query == "select * from REGION"
        assert import_job.output.type == AssetTypes.MLTABLE or import_job.output.type == AssetTypes.URI_FOLDER
        assert import_job.output.path == "azureml://datastores/workspaceblobstore/paths/output_dir/"

        import_job_3 = client.jobs.get(import_job.name)
        assert import_job_3.status in (JobStatus.CANCEL_REQUESTED, JobStatus.CANCELED, JobStatus.FAILED)

    @pytest.mark.skip("Skip for not ready.")
    @pytest.mark.e2etest
    def test_import_pipeline_submit_cancel(self, client: MLClient) -> None:
        pipeline: PipelineJob = load_job("./tests/test_configs/import_job/import_pipeline_test.yml")
        self.validate_test_import_pipepine_submit_cancel(pipeline, client, is_dsl=False)

    @pytest.mark.skip("Skip for not ready.")
    @pytest.mark.e2etest
    def test_import_dsl_pipeline_submit_cancel(self, client: MLClient) -> None:
        def generate_dsl_pipeline():
            # 1. Load component funcs
            import_func = load_component("./tests/test_configs/import_job/import_component_test.yml")
            prep_func = load_component("./tests/test_configs/dsl_pipeline/nyc_taxi_data_regression/prep.yml")

            # 2. Construct pipeline
            @dsl.pipeline(compute="cpu-cluster", default_datastore="workspaceblobstore")
            def sample_pipeline():
                import_job = import_func(
                    type="azuresqldb", connection="azureml:my_username_password", query="select * from REGION"
                )
                prep_job = prep_func(raw_data=import_job.outputs.output)
                return {
                    "pipeline_job_imported_data": import_job.outputs.output,
                    "pipeline_job_prepped_data": prep_job.outputs.prep_data,
                }

            pipeline = sample_pipeline()
            pipeline.outputs.pipeline_job_imported_data.data = "/imported_data"
            pipeline.outputs.pipeline_job_prepped_data.data = "/prepped_data"
            pipeline.outputs.pipeline_job_prepped_data.mode = "rw_mount"
            return pipeline

        pipeline = generate_dsl_pipeline()
        self.validate_test_import_pipepine_submit_cancel(pipeline, client, is_dsl=True)

    @classmethod
    def validate_test_import_pipepine_submit_cancel(cls, pipeline: PipelineJob, client: MLClient, is_dsl: bool) -> None:
        import_pipeline: PipelineJob = assert_job_cancel(job=pipeline, client=client)

        import_step = "import_job" if is_dsl else "import_step"
        assert import_step in import_pipeline.jobs
        assert isinstance(import_pipeline.jobs[import_step], Import)
        assert import_pipeline.jobs[import_step].type == JobType.IMPORT
        assert import_pipeline.jobs[import_step].inputs["type"]._data == "azuresqldb"
        assert import_pipeline.jobs[import_step].inputs["connection"]._data == "azureml:my_username_password"
        assert import_pipeline.jobs[import_step].inputs["query"]._data == "select * from REGION"

        if not is_dsl:
            assert (
                import_pipeline.jobs[import_step].outputs["output"].path
                == "azureml://datastores/workspaceblobstore/paths/output_dir/"
            )
            assert import_pipeline.jobs[import_step].outputs["output"]._data.type == AssetTypes.MLTABLE
            assert (
                import_pipeline.jobs[import_step].outputs["output"]._data.path
                == import_pipeline.jobs["import_step"].outputs["output"].path
            )

        import_pipeline_2 = client.jobs.get(import_pipeline.name)
        assert import_step in import_pipeline_2.jobs
        assert isinstance(import_pipeline_2.jobs[import_step], Import)
        assert import_pipeline_2.jobs[import_step].type == import_pipeline.jobs[import_step].type
        assert import_pipeline_2.jobs[import_step].compute == "DataFactory"
        assert (
            import_pipeline_2.jobs[import_step].inputs["type"]._data
            == import_pipeline.jobs[import_step].inputs["type"]._data
        )
        assert (
            import_pipeline_2.jobs[import_step].inputs["connection"]._data
            == import_pipeline.jobs[import_step].inputs["connection"]._data
        )
        assert (
            import_pipeline_2.jobs[import_step].inputs["query"]._data
            == import_pipeline.jobs[import_step].inputs["query"]._data
        )

        if not is_dsl:
            assert (
                import_pipeline_2.jobs[import_step].outputs["output"].path
                == import_pipeline.jobs[import_step].outputs["output"].path
            )
            assert (
                import_pipeline_2.jobs[import_step].outputs["output"]._data.type
                == import_pipeline.jobs[import_step].outputs["output"]._data.type
            )
            assert (
                import_pipeline_2.jobs[import_step].outputs["output"]._data.path
                == import_pipeline.jobs[import_step].outputs["output"]._data.path
            )

        import_pipeline_3 = client.jobs.get(import_pipeline.name)
        assert import_pipeline_3.status in (JobStatus.CANCEL_REQUESTED, JobStatus.CANCELED)

    @pytest.mark.e2etest
    def test_import_job_download(self, randstr: Callable[[str], str], client: MLClient) -> None:
        job = client.jobs.create_or_update(
            load_job(
                "./tests/test_configs/import_job/import_job_test.yml",
                params_override=[{"name": randstr("name")}],
            )
        )

        wait_until_done(client=client, job=job)

        with TemporaryDirectory() as tmp_dirname:
            tmp_path = Path(tmp_dirname)
            client.jobs.download(name=job.name, download_path=tmp_path, all=True)

            best_child_run_artifact_dir = tmp_path / "artifacts"

            assert best_child_run_artifact_dir.exists()
            assert next(best_child_run_artifact_dir.iterdir(), None), "No artifacts for child run were downloaded"
